"""
Tools to enhance compatability with DataJoint.
"""

import inspect

import datajoint as dj
from datajoint.user_tables import UserTable

from .base import Base
from .logging import getLogger
from .user_tables import Computed, Imported, Lookup, Manual, Part

logger = getLogger(__name__)

djp_mapping = {
    'Lookup': Lookup,
    'Manual': Manual,
    'Computed': Computed,
    'Imported': Imported,
    'Part': Part
}

def add_datajoint_plus(module):
    """
    Adds DataJointPlus recursively to DataJoint tables inside the module.
    """
    
    for name in dir(module):
        try:
            if name in ['key_source', '_master', 'master']:
                continue
            obj = getattr(module, name)
            if inspect.isclass(obj) and issubclass(obj, UserTable) and not issubclass(obj, Base):
                bases = []
                for b in obj.__bases__:
                    if issubclass(b, UserTable):
                        b = djp_mapping[b.__name__]
                    bases.append(b)
                obj.__bases__ = tuple(bases)
                obj.parse_hash_info_from_header()
                add_datajoint_plus(obj)
        except:
            logger.exception(f'Could not add DataJointPlus to {name}.')


def reassign_master_attribute(module):
    """
    Overwrite .master attribute in DataJoint part tables to map to master class from current module. This is required if the DataJoint table is inherited.
    """
    for name in dir(module):
        # Get DataJoint tables
        if inspect.isclass(getattr(module, name)) and issubclass(getattr(module, name), dj.Table):
            obj = getattr(module, name)
            for nested in dir(obj):
                # Get Part tables
                if inspect.isclass(getattr(obj, nested)) and issubclass(getattr(obj, nested), dj.Part):
                    setattr(getattr(obj, nested), '_master', obj)





import re
import inspect
from .logging import getLogger
import datajoint as dj

from datajoint_plus.utils import unwrap

logger = getLogger(__name__)

class StrToTable:
    """
    Abstract class -1
    """ 
    def __init__(self, source, context=None, **kwargs):
        self.source = source
        self.declaration_context = context or inspect.currentframe().f_back.f_globals
        self.kwargs = kwargs
        for k, v in kwargs.items():
            setattr(self, k, v)

    @property
    def table(self):
        try:
            table = eval(self.source, self.declaration_context)
            if inspect.isclass(table):
                return table()
            else:
                return table
        except:
            msg = f'Unable to instantiate {self.__class__.__qualname__}.'
            logger.exception(msg)
            raise NotImplementedError(msg)
    
    @property
    def base(self):
        b = self.table
        try:
            while not isinstance(b, dj.user_tables.UserTable):
                b = getattr(b, '_arg')
        except:
            msg = f'Unable to extract base table of type {self.table.__class__.__qualname__}.'
            logger.error(msg)
            raise NotImplementedError(msg)
        return b
    
    def __repr__(self):
        return f'{self.__class__.__qualname__}({self.source})'
    
    def __call__(self):
        return self.table

class Base:
    def __init__(self, source, type):
        if issubclass(source.__class__, type):
            self.__dict__ = source.__dict__
        else:
            source = unwrap(source)
            assert isinstance(source, str), f'Source must be of type {type.__name__} or str, not {source.__class__.__name__}.'
            self.source = source

class Definition:
    def __init__(self, source):
        if issubclass(source.__class__, __class__):
            self.__dict__ = source.__dict__
    
    def make_definition(self):
        for d in self._definition:
            foreign_key = {'primary': [], 'secondary': []}
            for ps in ['primary', 'secondary']:
                for emd_type in ['entities', 'methods', 'destinations']:
                    emd_nt = getattr(self, emd_type)
                    for emd in emd_nt:
                        if emd.inheritance == ps:
                            foreign_key[ps].extend(f"-> self.str_to_base(**{{'emd_type': '{emd_type}', 'source': '{emd.source}', 'context': self.declaration_context}}) \n") 
            
            definition = ''.join([
                "-> master \n",
                "".join(foreign_key['primary']),
                """
                ---
                """,
                "".join(foreign_key['secondary'])
            ])
            if self._append_timestamp_to_definition:
                definition += f"""
                {self._timestamp_name}=CURRENT_TIMESTAMP: timestamp # 
                """
            
            # format definition
            f_definition = []
            for line in re.split("\n", definition):
                line = line.strip(' ')
                if line:
                    f_definition.append(line)
            self.definition = '\n '.join(f_definition)
        
        return definition

    @property
    def definition(self):
        return self.make_definition()

    def __call__(self):
        return self.definition()

    class ForeignKey(Base):
        def __init__(self, source):
            super().__init__(source, __class__)

    
    class Attribute(Base):
        pass

    
    # Aliases
    Fk = ForeignKey
    At = Attribute



"""
Extensions of DataJoint Table
"""

from .logging import getLogger

import datajoint as dj

from datajoint_plus.hash import generate_table_id

from .utils import classproperty, goto
from .version import __version__ as version

logger = getLogger(__name__)


class Table(dj.table.Table):
    """
    Extensions to DataJoint Table
    """
    
    _table_log_ = None

    def declare(self, context=None):
        super().declare(context=context)
        if not isinstance(self, TableLog):
            self._table_log(self.table_id, self.full_table_name, action='add')
    
    def drop(self):
        dj.table.Table.drop(self)
        self._table_log(self.table_id, action='delete')

    def drop_quick(self):
        dj.table.Table.drop_quick(self)
        self._table_log(self.table_id, action='delete')

    def goto(self, table_id=None, full_table_name=None, tid_attr=None, ftn_attr=None, directory='__main__', return_free_table=False, ftn_lookup=None):
        """
        Extends datajoint_plus.utils.goto with the additional options:
            :param tid_attr: (str) default=None, attribute in self that points to a table_id. Will be used as such: table_id = self.fetch1(tid_attr)
            :param ftn_attr: (str) default=None, attribute in self that points to a full_table_name. Will be used as such: full_table_name = self.fetch1(ftn_attr)
            :param return_free_table: (bool) Provides a FreeTable generated from full_table_name instead of using goto. 
                If only table_id is known, ftn_lookup should provide a mapping between table_id and full_table_name. 
            :param ftn_lookup: (DJ table) DataJoint table to query table_id and retrieve full_table_name if return_free_table is True
                By default the following query will be attempted: (ftn_lookup & {'table_id': table_id}).fetch1('full_table_name')
                If ftn_lookup uses a different attr name for table_id, provide it to tid_attr.
        """
        if isinstance(self, str):
            raise AttributeError('Instantiate table to run goto().')
        
        if len(self)==1:
            try:
                table_id = self.fetch1('table_id')
            except:
                pass
        
        if table_id is None and tid_attr is not None:
            table_id = self.fetch1(tid_attr)
        
        if full_table_name is None and ftn_attr is not None:
            full_table_name = self.fetch1(ftn_attr)

        if full_table_name is None and table_id is None:
            raise Exception('Provide either table_id/full_table_name or tid_attr/ ftn_attr.')

        if return_free_table:
            if full_table_name is None:
                try:
                    if tid_attr is None:
                        tid_attr = 'table_id'
                    full_table_name = (ftn_lookup & {tid_attr: table_id}).fetch1('full_table_name')
                except Exception as e:
                    raise Exception('If return_free_table=True and full_table_name is not provided, table_id and ftn_lookup must be provided.')
            return FreeTable(self.connection, full_table_name)
        
        else:
            return goto(table_id=table_id, full_table_name=full_table_name, directory=directory)

    @property
    def _table_log(self):
        if self._table_log_ is None:
            self._table_log_ = TableLog(self.connection, database=self.database)
        return self._table_log_

    @classproperty
    def table_id(cls):
        return generate_table_id(cls.full_table_name)


class FreeTable(Table, dj.FreeTable):
    """
    Extension to DataJoint FreeTable
    """
    
    def drop(self):
        dj.table.Table.drop(self)
        self._table_log(full_table_name = self.full_table_name, action='delete')

    def drop_quick(self):
        dj.table.Table.drop_quick(self)
        self._table_log(full_table_name = self.full_table_name, action='delete')


class TableLog(Table):
    """
    Log of tables in each schema.
    Instances are callable.  Calls with table hash return the table entry. 
    Calls with hash and action update the table.
    """

    def __init__(self, arg, database=None):
        super().__init__()

        if isinstance(arg, TableLog):
            # copy constructor
            self.database = arg.database
            self._connection = arg._connection
            self._definition = arg._definition
            self._user = arg._user
            return

        self.database = database
        self._connection = arg
        self._definition = f"""    # tables in `{database}`
        table_id        :varchar(32)  # unique hash of full_table_name
        ---
        full_table_name : varchar(450) # name of table
        exists          : tinyint  # 1 - table exists in schema; 0 - table no longer exists
        djp_version     : varchar(32)  # version of datajoint_plus used to generate table_id
        timestamp = CURRENT_TIMESTAMP : timestamp # timestamp of entry (not necessarily when table was created)
        """

        if not self.is_declared:
            self.declare()
            self.connection.dependencies.clear()
        self._user = self.connection.get_user()

    @property
    def definition(self):
        return self._definition

    @property
    def table_name(self):
        return '~tables'
    
    @property
    def table_id(self):
        return generate_table_id(self.full_table_name)
    
    def __call__(self, table_id=None, full_table_name=None, action=None, directory='__main__'):
        """
        :param table_id: unique ID of table to insert
        :param full_table_name: full table name (database + table)
        :param action: Options - 
            add - inserts new table to log 
            delete - deletes table from log
        """
        try:
            if action == 'add':
                assert full_table_name is not None, 'full_table_name needed to add table'
                
                if table_id is None:
                    table_id = generate_table_id(full_table_name)
                else:
                    assert table_id == generate_table_id(full_table_name), 'Provided table_id does not match generated table_id.'

                restr = self & {'table_id': table_id}

                if len(restr) == 0:
                    self.insert1(
                        dict(
                            table_id=table_id, 
                            full_table_name=full_table_name, 
                            exists=1,
                            djp_version=version
                            ),
                        )
                restr._update('exists', 1)
      
            else:
                table_id_restr = f'table_id LIKE "{table_id}%"' if table_id is not None else None
                full_table_name_restr = {'full_table_name': full_table_name} if full_table_name is not None else None
                
                restr = [r for r in [table_id_restr, full_table_name_restr] if r is not None]
                restr = self & dj.AndList(restr)

                if action == 'delete':
                    assert ~((table_id is None) and (full_table_name is None)), 'Provide table_id or full_table_name to delete.'
                    assert len(restr) == 1, 'There should be only one entry to delete.'
                    restr._update('exists', 0)
                    return

                if (table_id is None) and (full_table_name is None):
                    return self

                table_id, full_table_name = restr.fetch1('table_id', 'full_table_name')
                table = goto(table_id, directory=directory)
                return table

        except Exception as e:
            logger.exception(e)
            logger.info('failure interacting with ~tables')

    @property
    def exists(self):
        """Returns existing tables"""
        return FreeTable(self.connection, self.full_table_name) & dj.AndList([{'exists': 1}])

    @property
    def deleted(self):
        """Returns deleted tables"""
        return FreeTable(self.connection, self.full_table_name) & dj.AndList([{'exists': 0}])

    def delete(self):
        """bypass interactive prompts and cascading dependencies"""
        self.delete_quick()

    def drop(self):
        """bypass interactive prompts and cascading dependencies"""
        self.drop_quick()



import collections
import hashlib
import inspect

import pandas as pd
import simplejson
from datajoint.expression import QueryExpression

from .errors import ValidationError


def generate_hash(rows, add_constant_columns:dict=None):
    """
    Generates hash for provided rows. 

    :param rows (pd.DataFrame, dict): Rows to hash. `type(rows)` must be able to instantiate a pandas dataframe.
    :param add_constant_columns (dict):  Each key:value pair will be passed to the dataframe to be hashed as `df[k]=v`, adding a column length `len(df)` with name `k` and filled with values `v`. 

    :returns: md5 hash
    """
    df = pd.DataFrame(rows)
    if add_constant_columns is not None:
        assert isinstance(add_constant_columns, dict), f' arg add_constant_columns must be Python dictionary instance.'
        for k, v in add_constant_columns.items():
            df[k] = v
    # permutation invariant hashing
    df = df.sort_index(axis=1)
    df = df.sort_values(by=df.columns.tolist()) 
    encoded = simplejson.dumps(df.to_dict(orient='records')).encode()
    dhash = hashlib.md5()
    dhash.update(encoded)
    return dhash.hexdigest()


def _validate_rows_for_hashing(rows):
    """
    Validates rows for `generate_hash`.
    """
    validated = False
    if isinstance(rows, pd.DataFrame):
        pass
    elif (inspect.isclass(rows) and issubclass(rows, QueryExpression)) or isinstance(rows, QueryExpression):
        pass
    elif isinstance(rows, list) or isinstance(rows, tuple):
        for row in rows:
            assert isinstance(row, collections.abc.Mapping), 'Cannot hash attributes unless row attributes are named. Try inserting a pandas dataframe, a DataJoint expression or a list of dictionaries.'
        pass
    else:
        raise ValidationError('Format of rows not recognized. Try inserting a list of dictionaries, a DataJoint expression or a pandas dataframe.')

        
def validate_and_generate_hash(rows, **kwargs):
    """
    Generates hash for provided rows with row validation for DataJoint.

    :rows: see `generate_hash`
    :kwargs: passed to `generate_hash`

    :returns: hash from `generate_hash`
    """
    _validate_rows_for_hashing(rows)
    return generate_hash(rows, **kwargs)


def generate_table_id(full_table_name):
    """
    Generates table_id by hashing full_table_name.

    :param: full_table_name 
    :returns: table_id
    """
    return generate_hash([{'full_table_name': full_table_name}])


import logging
import logging.config
import os
from pathlib import Path
from datajoint_plus.config.logging.load_config import import_config_yaml_as_dict
from .config import config


def basicConfig(filename=None, level=None, format='%(asctime)s - %(name)s:%(levelname)s:%(message)s', datefmt="%m-%d-%Y %I:%M:%S %p %Z", force=True, **kwargs):
    if filename is not None:
        filename = Path(filename)
    
    level = level if level is not None else config['loglevel']

    numeric_level = getattr(logging, level.upper(), None)
    if not isinstance(numeric_level, int):
        raise ValueError('Invalid log level: %s' % level)

    logging.basicConfig(filename=filename, level=numeric_level, format=format, datefmt=datefmt, force=force, **kwargs)


def getLogger(name, config_file='console.yml', search_templates=True, level=None, update_root_level=True, **kwargs):
    """
    :param name: (str) name to assign to logging.getLogger object
    :param config_file: (str or Path) path to logconfig file  
    :param search_templates: (bool) whether to search djp templates for config files
    :param level: (str) logger level to overwrite config file
    :param update_root_level: (bool) if True, will update root logger level. 
        Only applicable if level is provided.
    :param kwargs: kwargs to pass as replacement_dict (see datajoint_plus.config.logging.load_config.import_config_yaml_as_dict)
    """
    config = import_config_yaml_as_dict(path_to_file=config_file, search_templates=search_templates, replacement_mapping=kwargs)
    
    # name logger
    config['loggers'][name] = config['loggers'].pop('unnamed')

    # instantiate logger
    logging.config.dictConfig(config)

    # get logger
    logger = logging.getLogger(name)

    # overwrite level
    if level is not None:
        logger.setLevel(level)
        if update_root_level:
            logging.getLogger().setLevel(level) # root logger
    
    return logger


class LogFileManager:
    def __init__(self, name, filename, base_dir=None, config_file=None, search_templates=True, level=None, **kwargs):
        """
        :param name: (str) name to assign to logging.getLogger object
        :param filename: (str or Path) path to log file relative to base_dir
        :param base_dir" (str or Path) base directory for log files
            if None - defaults to config['DJ_LOG_BASE_DIR']
        :param config_file: (str or Path) path to logconfig file  
        :param search_templates: (bool) whether to search djp templates for config files
        :param level: (bool) logger level to overwrite config file
        :param kwargs: kwargs to pass to getLogger 
        """
        self.name = name
        self.level = level
        self.base_dir = Path(os.getenv('DJ_LOG_BASE_DIR')) if base_dir is None else base_dir
        self.filename = Path(filename)
        self.filepath = self.base_dir.joinpath(self.filename)
        self.config_file = config_file
        self._search_templates = search_templates
        self.level = level
        self.kwargs = kwargs

        if not self.filepath.parent.exists():
            self.filepath.parent.mkdir(exist_ok=True, parents=True)

    def __repr__(self):
        return f'LogFileManager({self.name})'

    def __call__(self, method, *args, **kwargs):
        getattr(self.logger, method)(*args, **kwargs)

    @property
    def filepaths(self):
        """
        Returns a list of log files accessible by Logger.
        """
        return [p for p in self.filepath.parent.iterdir() if str(self.filepath.name) in p.name]

    @property
    def logger(self):
        self.kwargs['filename'] = self.filename
        return getLogger(
            self.name, 
            config_file=self.config_file, 
            search_templates=self._search_templates,
            level=self.level,
            **self.kwargs
        )

    def lines(self, log_file_ind=None, return_path=False):
        """
        Returns list of lines in specified log.

        :param log_file_ind: (int) optional, the index of the log file to access from self.paths
            default - None, same as 0
        :param return_path: (bool) if True, returns path of log file
        :returns: list of str
            if return_path is True, also returns path to log file
        """
        if log_file_ind is not None:
            try:
                path = self.filepaths[log_file_ind]
            except IndexError:
                print('No log file at specified index.')
                return
        else:
            path = self.filepath
            if not path.exists():
                print('Log file does not exist yet. Log an entry to create it.')
                return

        with open(path, 'r') as f:
            lines = f.readlines()
            if return_path:
                return lines, path
            else:
                return lines

    def _showlines(self, n_entries, log_file_ind):
        result = self.lines(log_file_ind, return_path=True)
        if result is None:
            return
        lines, path = result
        n_lines = len(lines)
        n_lines_shown = min(n_lines, n_entries)
        return lines, n_lines, n_lines_shown, path

    def head(self, n_entries:int=10, log_file_ind=None):
        """
        Prints log entries from the top of the file.

        :param n_entries: (int) the number of entries to return
        :param log_file_ind: (int) the index of the log file to access from self.paths

        :returns: None
        """
        result = self._showlines(n_entries, log_file_ind)
        if result is None:
            return
        lines, n_lines, n_lines_shown, path = result
        print(f'Showing first {n_lines_shown} entries from ".../{path.name}" out of {n_lines} total entries: \n')
        for line in lines[:n_entries]:
            line = line.strip('\n')
            print(line)

    def tail(self, n_entries:int=10, log_file_ind=None):
        """
        Prints log entries from the end of the file.

        :param n_entries: (int) the number of entries to return
        :param log_file_ind: (int) the index of the log file to access from self.paths

        :returns: None
        """
        result = self._showlines(n_entries, log_file_ind)
        if result is None:
            return
        lines, n_lines, n_lines_shown, path = result
        print(f'Showing last {n_lines_shown} entries from ".../{path.name}" out of {n_lines} total entries: \n')
        for line in lines[::-1][:n_entries][::-1]:
            line = line.strip('\n')
            print(line)


logger = getLogger(__name__)


"""
Tools to modify DataJoint table headings.
"""

import re

import numpy as np
import pandas as pd


def parse_definition(definition):
    """
    Parses DataJoint definitions. Extracts the following line types, where lines are separated by newlines `\\n`:
        - headers: a line that starts with hash `#`.
        - dependencies: a line that does not start with hash `#` and contains an arrow `->`.
        - attributes: a line that contains a colon `:`, and may contain a hash `#`, as long as the colon `:` precedes the hash `#`.
        - divider: a line that does not start with a hash `#` and contains a DataJoint divider `---`.
        - non-matches: a line that does not meet any of the above criteria. 
        
    :param definition: DataJoint table definition to be parsed.

    :returns: 
        parsed_inds (dict): dictionary of line types and matching line indexes. 
        parsed_contents (dict): a dictionary of line types and matching line contents.
        parsed_stats (dict): a dictionary of line types and a summary of the number of line matches for each type. 
    """

    lines = re.split('\n', definition) # seperate definition lines
    lines_without_spaces = [re.sub(' +', '', l) for l in lines]

    char_inds = []
    for l in lines_without_spaces:
        ind_dict = {}
        for name, char in zip(['hash', 'divider', 'arrow', 'colon'], ['#', '---', '->', ':']):
            ind_dict[name] = [m.start() for m in re.finditer(char, l)]
        char_inds.append(ind_dict)

    df = pd.DataFrame(char_inds)

    df['contains_hash'] = df.apply(lambda x: x.hash != [], axis=1)
    df['contains_divider'] = df.apply(lambda x: x.divider != [], axis=1)
    df['contains_arrow'] = df.apply(lambda x: x.arrow != [], axis=1)
    df['contains_colon'] = df.apply(lambda x: x.colon != [], axis=1)
    df['hash_pos_0'] = df.apply(lambda x: 0 in x.hash, axis=1)
    df['colon_before_hash'] = df.apply(lambda x: x.colon[0] < x.hash[0] if x.colon and x.hash else False, axis=1)

    header_query = "hash_pos_0 == True"
    dependency_query = "hash_pos_0 == False and contains_arrow == True"
    attribute_query = "(contains_colon==True and contains_hash==False) or colon_before_hash==True"
    divider_query = "hash_pos_0 == False and contains_divider == True"

    names = ['headers', 'dependencies', 'attributes', 'dividers']
    queries = [header_query, dependency_query, attribute_query, divider_query]

    parsed_inds = {}
    for name, query in zip(names, queries):
        q = df.query(query)
        df = df.drop(df.query(query).index.values)
        if len(q)> 0:
            parsed_inds[name] = q.index.values
        else:
            parsed_inds[name] = []
    parsed_inds['non-matches'] = df.index.values

    parsed_contents = {}
    parsed_stats = {}

    for n in names + ['non-matches']:
        parsed_contents[n] = [lines[i] for i in parsed_inds[n]]
        parsed_stats[n] = f'Found {len(parsed_inds[n])} line(s) matching the profile of {n}.'

    return parsed_inds, parsed_contents, parsed_stats


def reform_definition(parsed_inds, parsed_contents):
    """
    Reforms DataJoint definition after parsing by `parse_definition`. 

    :param parsed_inds (dict): see `parse_definition`
    :param parsed_contents (dict): see `parse_definition`

    :returns: DataJoint definition
    """
    n_lines = len(np.concatenate([i for i in parsed_inds.values()]))
    content_list = [''] * n_lines
    for ii, cc in zip(parsed_inds.values(), parsed_contents.values()):
        for i, c in zip(ii, cc):
            content_list[int(i)] = c

    definition = """"""
    for c in content_list:
        definition += c + """\n"""
    
    return definition



"""
Abstract classes for DataJointPlus
"""

from collections import Counter
import inspect
from pathlib import Path
import re
import traceback

import numpy as np
import datajoint as dj
from datajoint.table import QueryExpression, FreeTable
from .config import config

from .logging import LogFileManager, getLogger
from .enum import JoinMethod
from .errors import OverwriteError, ValidationError
from .hash import generate_hash
from .heading import parse_definition, reform_definition
from .utils import classproperty, format_rows_to_df, format_table_name, unwrap, wrap, load_dependencies
from .validation import (_is_overwrite_validated,
                         _validate_hash_name_type_and_parse_hash_len,
                         pairwise_disjoint_set_validation)

logger = getLogger(__name__)

class Base:
    _is_insert_validated = False
    _enable_table_modification = True

    # required for hashing
    enable_hashing = False
    hash_name = None
    hashed_attrs = None

    # hash params
    hash_group = False
    hash_table_name = False
    _hash_len = None

    # header params
    _add_info_to_header = True
    _add_class_name_to_header = True
    _add_hash_name_to_header = True
    _add_hashed_attrs_to_header = True
    _add_hash_params_to_header = True
    
    # logging
    loglevel = config['loglevel']

    @classmethod
    def _init_validation(cls, **kwargs):
        """
        Validation for initialization of subclasses of abstract class Base. 
        """
        for attr in ['enable_hashing', 'hash_group', 'hash_table_name', '_add_hash_name_to_header', '_add_hash_params_to_header', '_add_hashed_attrs_to_header']:
            assert isinstance(getattr(cls, attr), bool), f'"{attr}" must be boolean.'           

        for attr in ['hash_name', 'hashed_attrs']:
            assert not isinstance(getattr(cls, attr), bool), f'"{attr}" must not be boolean.'
        
        if cls.enable_hashing:
            for required in ['hash_name', 'hashed_attrs']:
                if getattr(cls, required) is None:
                    raise NotImplementedError(f'Hashing requires class to implement the property "{required}".')
        
        # ensure one attribute
        for name in ['hash_name']:
            attr = getattr(cls, name)
            if attr is not None:
                if isinstance(attr, list) or isinstance(attr, tuple):
                    if len(attr) > 1:
                        raise NotImplementedError(f'Only one attribute allowed in "{name}".')
                    else:
                        attr = attr[0]

        # ensure "hashed_attrs" wrapped in list or tuple
        if cls.hashed_attrs is not None:
            if not isinstance(cls.hashed_attrs, list) and not isinstance(cls.hashed_attrs, tuple):
                cls.hashed_attrs = [cls.hashed_attrs]
            else:
                cls.hashed_attrs = cls.hashed_attrs

        # ensure sets are disjoint
        cls._must_be_disjoint = {}
        for name in ['hash_name', 'hashed_attrs']:
            attr = getattr(cls, name)
            if attr is not None:
                if not isinstance(attr, list) and not isinstance(attr, tuple):
                    cls._must_be_disjoint[name] = set([attr])
                else:
                    cls._must_be_disjoint[name] = set(attr)
        pairwise_disjoint_set_validation(list(cls._must_be_disjoint.values()), list(cls._must_be_disjoint.keys()), error=NotImplementedError)

        # set kwarg defaults
        kwargs.setdefault('hash_part_table_names', None)

        # modify header
        hash_info_dict = dict(
            add_class_name = cls._add_class_name_to_header,
            hash_name=cls.hash_name if cls._add_hash_name_to_header else None,
            hashed_attrs=cls.hashed_attrs if cls._add_hashed_attrs_to_header else None,
            hash_group=True if cls.hash_group and cls._add_hash_params_to_header else None, # only add if set to True (default is False)
            hash_table_name=True if cls.hash_table_name and cls._add_hash_params_to_header else None, # only add if set to True (default is False)
            hash_part_table_names=False if kwargs['hash_part_table_names'] is False and cls._add_hash_params_to_header else None # only add if set to False (default is True)
        )
        if cls._add_info_to_header:
            cls._modify_header(**hash_info_dict)
    
    @classproperty
    def class_name(cls):
        return cls.__qualname__

    @classproperty
    def hash_len(cls):
        if cls.hash_name is not None and cls._hash_len is None:
            cls._hash_name_validation()
        return cls._hash_len

    @classmethod
    def _insert_validation(cls):
        """
        Validation for insertion to DataJoint tables that are subclasses of abstract class Base. 
        """
        # ensure sets are disjoint
        pairwise_disjoint_set_validation(list(cls._must_be_disjoint.values()), list(cls._must_be_disjoint.keys()), error=AttributeError)


        # ensure "index" not in attributes
        if "index" in cls.heading.names:
            raise AttributeError(f'Attributes cannot be named "index". There is a bug in this DJ version that does not handle this keyword correctly with respect to MySQL.')

        cls._is_insert_validated = True

    @classmethod
    def load_dependencies(cls, force=True, verbose=True):
        """
        Loads dependencies into DataJoint networkx graph. 
        """
        load_dependencies(cls.connection, force=force, verbose=verbose)


    @classmethod
    def add_constant_attrs_to_rows(cls, rows, constant_attrs:dict={}, overwrite_rows=False):
        """
        Adds attributes to all rows.
        Warning: rows must be able to be safely converted into a pandas dataframe.
        :param rows (pd.DataFrame, QueryExpression, list, tuple): rows to pass to DataJoint `insert`. 
        :param constant_attrs (dict): Python dictionary to add to every row in rows
        :overwrite_rows (bool): Whether to overwrite key/ values in rows. If False, conflicting keys will raise a ValidationError.
        :returns: modified rows
        """   
        assert isinstance(constant_attrs, dict), 'constant_attrs must be a dict'

        rows = format_rows_to_df(rows)

        for k, v in constant_attrs.items():
            if _is_overwrite_validated(k, rows, overwrite_rows):
                rows[k] = v

        return rows

    @classmethod
    def include_attrs(cls, *args):
        """
        Returns a projection of cls that includes only the attributes passed as args.

        Note: The projection is NOT guaranteed to have unique rows, even if it contains only primary keys. 
        """
        return cls.proj(..., **{a: '""' for a in cls.heading.names if a not in args}).proj(*[a for a in cls.heading.names if a in args])
    
    @classmethod
    def exclude_attrs(cls, *args):
        """
        Returns a projection of cls that excludes all attributes passed as args. 
        
        Note: The projection is NOT guaranteed to have unique rows, even if it contains only primary keys. 
        """
        return cls.proj(..., **{a: '""' for a in cls.heading.names if a in args}).proj(*[a for a in cls.heading.names if a not in args])
             
    @classmethod
    def hash1(cls, rows, unique=False, as_dict=False, **kwargs):
        """
        Hashes rows and requires a single hash as output.
        
        Warning: rows must be able to be safely converted into a pandas dataframe.
        kwargs: args for `add_hash_to_rows`
        
        :returns (str): hash
        """
        hashes = cls.hash(rows, unique=unique, as_dict=as_dict, **kwargs)
        assert len(hashes) == 1, 'Multiple hashes found. hash1 must return only 1 hash.'
        return unwrap(hashes)

    @classmethod
    def hash(cls, rows, unique=False, as_dict=False, **kwargs):
        """
        Hashes rows.
        Warning: rows must be able to be safely converted into a pandas dataframe.
        
        kwargs: args for `add_hash_to_rows`
        :param rows: rows containing attributes to be hashed. 
        :unique: If True, only unique hashes will be returned. If False, all hashes returned. 
        
        returns (list): list with hash(es)
        """ 
        rows = cls.add_hash_to_rows(rows, overwrite_rows=True) # perform hashing
        rows = getattr(rows, cls.hash_name) # get hash name
        rows = rows.values if not unique else rows.unique() 
        return rows.tolist() if not as_dict else [{cls.hash_name: i} for i in rows]

    @classmethod
    def restrict_with_hash(cls, hash, hash_name=None):
        """
        Returns table restricted with hash. 

        :param hash: hash to restrict with
        :param hash_name: name of attribute that contains hashes. 
            If hash_name is not None:
                Will use hash_name instead of cls.hash_name
            If hash_name is None:
                Will use cls.hash_name or raise ValidationError if cls.hash_name is None

        :returns: Table restricted with hash. 
        """
        if hash_name is None and hasattr(cls, 'hash_name'):
            hash_name = cls.hash_name
        
        if hash_name is None:
            raise ValidationError('Table does not have "hash_name" defined, provide it to restrict with hash.')
            
        return cls & {cls.hash_name: hash}

    @classmethod
    def _modify_header(cls, add_class_name=True, **kwargs):
        """
        Modifies definition header using a parseable syntax
        """
        # set defaults
        kwargs.setdefault('hash_name', None)
        kwargs.setdefault('hashed_attrs', None)
        kwargs.setdefault('hash_group', None)
        kwargs.setdefault('hash_table_name', None)
        kwargs.setdefault('hash_part_table_names', None)


        if hasattr(cls, 'definition') and isinstance(cls.definition, str):
            inds, contents, _ = parse_definition(cls.definition)
            headers = contents['headers']

            if len(headers) >= 1:
                header = headers[0]
                header = header.strip(' ').strip('#').strip(' ')
                cls.comment = header
            else:
                # create header
                header = """#"""
                cls.comment = ''

            if add_class_name:
                header = f"#~{cls.class_name} | " + header[header.find("#")+1:]

            # append hash info to header
            for attr in ['hash_name', 'hash_group', 'hash_table_name', 'hash_part_table_names']:
                if kwargs[attr] is not None:
                    header += f" | {attr}: {kwargs[attr]} " 

            if kwargs['hashed_attrs'] is not None:
                header += f" | hashed_attrs: "
                for i, h in enumerate(kwargs['hashed_attrs']):
                    header += f"{h}, " if i+1 < len(kwargs['hashed_attrs']) else f"{h} "
            
            # remove spaces from missing comment
            for n in ['||', '| |', '|  |']:
                if header.find(n):
                    header = header.replace(n, '|')

            try:
                # replace existing header with modified header
                contents['headers'][0] = header
        
            except IndexError:
                # add header
                contents['headers'].extend([header])
                
                # header should go before any dependencies or attributes
                header_ind = np.min(np.concatenate([inds['dependencies'], inds['attributes']]))
                inds['headers'].extend([header_ind])
                
                # slide index over 1 to accommodate new header
                for n in [k for k in inds.keys() if k not in ['headers']]:
                    if len(inds[n])>0:
                        inds[n][inds[n] >= header_ind] += 1
            
            # reform and set definition
            cls.definition = reform_definition(inds, contents)

    @classmethod
    def parse_hash_info_from_header(cls):
        """
        Parses hash_name and hashed_attrs from DataJoint table header and sets properties in class. 
        """
        msg = 'Could not parse % from header.'
        header = cls.heading.table_info['comment']
                
        # parse class name
        class_name_parse = re.findall('~(.*?)\|', header)
        if class_name_parse:
            class_name_matches = re.findall('\w+', unwrap(class_name_parse))
            try:
                if len(class_name_matches)==1:
                    cls.class_name = class_name_matches[0]
                elif len(class_name_matches)==2:
                    cls.class_name = '.'.join(class_name_matches)
                else:
                    raise('parsed class_name did not match expected format. ')
            except:
                logger.exception(msg, 'class_name')
        else:
            cls.class_name = ''
        
        if cls.class_name:
            # remove parsed class_name
            header = header.replace('~' + unwrap(class_name_parse) + '|', '')
        
        # parse hash attributes
        matches = re.split(r'\|', header)
        if matches:
            for match in matches:
                result = re.findall('\w+', match)
                if result:
                    parseable = ['hash_name', 'hashed_attrs', 'hash_group', 'hash_table_name', 'hash_part_table_names']
                    for attr in parseable:
                        if result[0] == attr:
                            try:
                                if attr == 'hash_name':
                                    cls.hash_name = result[1]
                                    header = header.replace(match, '') # remove match
                                if attr == 'hashed_attrs':
                                    cls.hashed_attrs = result[1:]
                                    header = header.replace(match, '')

                                if result[1] == 'True' or result[1] == 'False':
                                    setattr(cls, attr, eval(result[1]))
                                    header = header.replace(match, '')
                            except:
                                logger.exception(msg, attr)
        
        cls.comment = header.strip(' ').strip('|').strip(' ')


    @classmethod
    def add_hash_to_rows(cls, rows, overwrite_rows=False):
        """
        Adds hash to rows. 
        
        Warning: rows must be able to be safely converted into a pandas dataframe.
        :param rows (pd.DataFrame, QueryExpression, list, tuple): rows to pass to DataJoint `insert`.
        :overwrite_rows (bool): Whether to overwrite key/ values in rows. If False, conflicting keys will raise a ValidationError. 
        :returns: modified rows
        """
        assert cls.hashed_attrs is not None, 'Table must have hashed_attrs defined. Check if hashing was enabled for this table.'

        hash_table_name = True if cls.hash_table_name or (issubclass(cls, dj.Part) and hasattr(cls.master, 'hash_part_table_names') and getattr(cls.master, 'hash_part_table_names')) else False

        if hash_table_name:
            table_id = {'table_id': cls.table_id}
        else:
            table_id = None
            
        rows = format_rows_to_df(rows)

        for a in cls.hashed_attrs:
            assert a in rows.columns.values, f'hashed_attr "{a}" not in rows. Row names are: {rows.columns.values}'

        if _is_overwrite_validated(cls.hash_name, rows, overwrite_rows):
            rows_to_hash = rows[[*cls.hashed_attrs]]

            if cls.hash_group:
                rows[cls.hash_name] = generate_hash(rows_to_hash, add_constant_columns=table_id)[:cls.hash_len]

            else:
                rows[cls.hash_name] = [generate_hash([row], add_constant_columns=table_id)[:cls.hash_len] for row in rows_to_hash.to_dict(orient='records')]
                
        return rows

    @classmethod
    def _prepare_insert(cls, rows, constant_attrs, overwrite_rows=False, skip_hashing=False):
        """
        Prepares rows for insert by checking if table has been validated for insert, adds constant_attrs and performs hashing. 
        """
        
        if not cls._is_insert_validated:
            cls._insert_validation()
        
        if constant_attrs != {}:
            rows = cls.add_constant_attrs_to_rows(rows, constant_attrs, overwrite_rows)

        if cls.enable_hashing and not skip_hashing:
            try:
                rows = cls.add_hash_to_rows(rows, overwrite_rows=overwrite_rows)

            except OverwriteError as err:
                new = err.args[0]
                new += ' Or, to skip the hashing step, set skip_hashing=True.'
                raise OverwriteError(new) from None

        return rows

    def get(self, key={}, attrs={}):
        """
        Wrapper around fetch that can take a key to restrict self. 
        Results always return as a dictionary.

        :param key: (dict, QueryExpression, AndList, etc) a restriction for table
        :param attrs: (str or list/ tuple) A single attr can be provided as a str, or a list/ tuple of strings

        :returns: (dict) Dictionary containing fetch1 results
        """
        try:
            if attrs != {}:
                attrs = wrap(attrs)
                result = wrap((self & key).fetch1(*attrs))
                return {a: r for a, r in zip(attrs, result)}
            else:
                return (self & key).fetch1()
        except AttributeError as e:
            raise AttributeError(e.args[0] + f'. Did you instantiate the class?') from None

    @classmethod
    def _get_attr_name_from_type(cls, attr_type):
        """
        Gets the name(s) of the attribute of type attr_type.

        :attr_type: type of attribute to search. Corresponds to DataJoint attribute types.

        :returns: (list) list of attr_names matching attr_type
        """
        return [k for k, v in cls.heading.attributes.items() if v.type == attr_type]

    @classmethod
    def aggr_min(cls, attr_name:str):
        """
        Returns table restricted by the maximum entry of attr_name.

        :param attr_name: (str) name of attribute to aggregate over

        :returns: restricted dj_table
        """
        return cls & (dj.U(attr_name) * dj.U().aggr(cls, **{attr_name: f'min({attr_name})'}))

    @classmethod
    def aggr_max(cls, attr_name:str):
        """
        Returns table restricted by the minimum entry of attr_name.

        :param attr_name: (str) name of attribute to aggregate over

        :returns: restricted dj_table
        """
        return cls & (dj.U(attr_name) * dj.U().aggr(cls, **{attr_name: f'max({attr_name})'}))

    @classmethod
    def aggr_nunique(cls, attr_name:str):
        """
        Returns table restricted by the minimum entry of attr_name.

        :param attr_name: (str) name of attribute to aggregate over

        :returns: restricted dj_table
        """
        return dj.U().aggr(cls, **{attr_name: f'count(distinct {attr_name})'}).fetch1(attr_name)

    @classmethod
    def _timestamp_attr_validation(cls, ts_name=None):
        """
        Validates that ts_name is a timestamp attribute if provided, otherwise, checks that onle one timestamp attribute exists in the table.
        """
        if ts_name is None:
            ts_name = cls._get_attr_name_from_type('timestamp')
            if isinstance(ts_name, list) and len(ts_name) == 0:
                err = 'timestamp attribute not found.'
                logger.error(err)
                raise AttributeError(err)
            elif isinstance(ts_name, list) and len(ts_name) > 1:
                err = 'Multiple timestamp attributes found. Provide desired timestamp to ts_name'
                logger.error(err)
                raise AttributeError(err)
        else:
            if cls.heading.attributes[ts_name].type != 'timestamp':
                err = f'ts_name "{ts_name}" is not a timestamp.'
                logger.error(err)
                raise AttributeError(err)
        return unwrap(ts_name)

    @classmethod
    def get_earliest_entries(cls, ts_name=None):
        """
        Returns table restricted to the earliest entries. Requires table to have timestamp attribute.

        :param ts_name: (str) Name of timestamp attribute. 
            If None: searches heading for timestamp attribute
        
        :returns: restricted dj_table
        """
        return cls.aggr_min(cls._timestamp_attr_validation(ts_name))

    @classmethod
    def get_latest_entries(cls, ts_name=None):
        """
        Returns table restricted to the latest entries. Requires table to have timestamp attribute.

        :param ts_name: (str) Name of timestamp attribute. 
            If None: searches heading for timestamp attribute
        
        :returns: (dj_table) Restricted table
        """
        return cls.aggr_max(cls._timestamp_attr_validation(ts_name))
        
    @classproperty
    def Log(cls):
        return LogFileManager(
            name=cls.__module__ + '.' + cls.__qualname__, 
            filename=Path(cls.database).joinpath(cls.__qualname__ + '.log'), 
            config_file='filehandler_timed_rotating.yml',
            level=cls.loglevel
        )
        

class BaseMaster(Base):
    hash_part_table_names = True
    _is_hash_name_validated = False

    @classmethod
    def _init_validation(cls, **kwargs):
        """
        Validation for initialization of subclasses of abstract class BaseMaster. 
        """
        for attr in ['hash_table_name', 'hash_part_table_names']:
            assert isinstance(getattr(cls, attr), bool), f'"{attr}" must be a boolean.'

        super()._init_validation(hash_table_name=cls.hash_table_name, hash_part_table_names=cls.hash_part_table_names)

    @classmethod
    def _insert_validation(cls):
        """
        Validation for insertion into subclasses of abstract class BaseMaster. 
        """
        if cls.hash_name is not None:
            if cls.hash_name not in cls.heading.names:
                raise ValidationError(f'hash_name "{cls.hash_name}" must be present in table heading.')

            # hash_name validation
            if not cls._is_hash_name_validated:
                cls._hash_name_validation()
        
        super()._insert_validation()
    
    @classmethod
    def _hash_name_validation(cls):
        """
        Validates hash_name and sets hash_len
        """       
        cls._hash_len = _validate_hash_name_type_and_parse_hash_len(cls.hash_name, cls.heading.attributes)
        cls._is_hash_name_validated = True

    @classmethod
    def parts(cls, as_objects=False, as_cls=False, reload_dependencies=False):
        """
        Wrapper around Datajoint function `parts` that enables returning parts as part_names, objects, or classes, and enables reloading of Datajoint networkx graph dependencies.

        :param as_objects: 
            If True, returns part tables as objects
            If False, returns part table names
        :param as_cls:
            If True, returns part table classes (will override as_objects)
        :param reload_dependencies: 
            If True, will force reload Datajoint networkx graph dependencies. 

        :returns: list
        """
        cls.load_dependencies(force=reload_dependencies)

        cls_parts = [getattr(cls, d) for d in dir(cls) if inspect.isclass(getattr(cls, d)) and issubclass(getattr(cls, d), dj.Part)]
        for cls_part in [p.full_table_name for p in cls_parts]:
            if cls_part not in super().parts(cls):
                logger.warning('Part table defined in class definition not found in DataJoint graph. Reload dependencies.')

        if not as_cls:
            return super().parts(cls, as_objects=as_objects)
        else:
            return cls_parts

    @classmethod
    def number_of_parts(cls, reload_dependencies=False):
        """
        Returns the number of part tables belonging to cls. 
        """
        return len(cls.parts(reload_dependencies=reload_dependencies))

    @classmethod
    def has_parts(cls, reload_dependencies=False):
        """
        Returns True if cls has part tables. 
        """
        return cls.number_of_parts(reload_dependencies=reload_dependencies) > 0

    @classmethod
    def _format_parts(cls, parts):
        """
        Formats the part tables in arg parts. 
        """
        if not isinstance(parts, list) and not isinstance(parts, tuple):
            parts = [parts]
        
        new = []
        for part in parts:
            if inspect.isclass(part) and issubclass(part, dj.Part):
                new.append(part()) # instantiate if a class

            elif isinstance(part, dj.Part):
                new.append(part)

            elif isinstance(part, QueryExpression) and not isinstance(part, dj.Part):
                raise ValidationError(f'Arg "{part.full_table_name}" is not a valid part table.')

            else:
                raise ValidationError(f'Arg "{part}" must be a part table or a list or tuple containing one or more part tables.')

        return new

    @classmethod
    def union_parts(cls, part_restr={}, include_parts=None, exclude_parts=None, filter_out_len_zero=False, reload_dependencies=False):
        """
        Returns union of part table primary keys after optional restriction. Requires all part tables in union to have identical primary keys. 

        :params: see `restrict_parts`.

        :returns: numpy array object
        """  
        return np.sum([p.proj() for p in cls.restrict_parts(part_restr=part_restr, include_parts=include_parts, exclude_parts=exclude_parts, filter_out_len_zero=filter_out_len_zero, reload_dependencies=reload_dependencies)])

#     @classmethod
#     def keys_not_in_parts(cls, part_restr={}, include_parts=None, exclude_parts=None, master_restr={}, parts_kws={}):
#         return (cls & master_restr) - cls.union_parts(include_parts=include_parts, exclude_parts=exclude_parts, part_restr=part_restr, parts_kws=parts_kws)

    @classmethod
    def join_parts(cls, part_restr={}, join_method=None, join_with_master=False, include_parts=None, exclude_parts=None, filter_out_len_zero=False, reload_dependencies=False):
        """
        Returns join of part tables after optional restriction. 

        :params part_restr, include_parts, exclude_parts: see `restrict_parts`.
        :param join_method (str):
            - 'primary_only' - will project out secondary keys and will only join on primary keys
            - 'rename_secondaries' - will add the part table to all secondary keys
            - 'rename_collisions' - will add the part table name to secondary keys that are present in more than one part table
            - 'rename_all' - will add the part table name to all primary and secondary keys

        :param join_with_master (bool): If True, parts will be joined with cls before returning. 

        :returns: numpy array object
        """
        parts = cls.restrict_parts(part_restr=part_restr, include_parts=include_parts, exclude_parts=exclude_parts, filter_out_len_zero=filter_out_len_zero, reload_dependencies=reload_dependencies)
        
        if join_with_master:
            parts = [FreeTable(cls.connection, cls.full_table_name)] + parts

        collisions = None
        if join_method is None:
            try:
                return np.product(parts)

            except:
                traceback.print_exc()
                msg = 'Join unsuccessful. Try one of the following: join_method = '
                for i, j in enumerate(JoinMethod):
                    msg += f'"{j.value}", ' if i+1 < len(JoinMethod) else f'"{j.value}".'
                print(msg)
                return
        
        elif join_method == JoinMethod.PRIMARY.value:
            return np.product([p.proj() for p in parts])
        
        elif join_method == JoinMethod.SECONDARY.value:
            attributes_to_rename = [p.heading.secondary_attributes for p in parts]
            
        elif join_method == JoinMethod.COLLISIONS.value:
            attributes_to_rename = [p.heading.secondary_attributes for p in parts]
            collisions = [item for item, count in Counter(np.concatenate(attributes_to_rename)).items() if count > 1]
            
        elif join_method == JoinMethod.ALL.value:
            attributes_to_rename = [list(p.heading.attributes.keys()) for p in parts]

        else:
            msg = f'join_method "{join_method}" not implemented. Available methods: '
            for i, j in enumerate(JoinMethod):
                msg += f'"{j.value}", ' if i+1 < len(JoinMethod) else f'"{j.value}".'
            raise NotImplementedError(msg)

        renamed_parts = []
        for p, attrs in zip(parts, attributes_to_rename):
            if isinstance(p, dj.Part):
                name = format_table_name(p.table_name, snake_case=True, part=True).split('.')[1]
            else:
                name = format_table_name(p.table_name, snake_case=True)

            if collisions is not None:
                renamed_attribute = {name + '_' + a : a for a in attrs if a in collisions}
            else:
                renamed_attribute = {name + '_' + a : a for a in attrs}
            renamed_parts.append(p.proj(..., **renamed_attribute))
            
        return np.product(renamed_parts)
    
    @classmethod
    def restrict_parts(cls, part_restr={}, include_parts=None, exclude_parts=None, filter_out_disjoint=False, filter_out_len_zero=False, reload_dependencies=False):
        """
        Restricts part tables of cls. 

        :param part_restr: restriction to restrict part tables with.
        :param include_parts: (part table or list of part tables) part table(s) to restrict. If None, will restrict all part tables of cls.
        :param exclude_parts: (part table or list of part tables) part table(s) to exclude from restriction.
        :param filter_out_disjoint: (bool) If True, included parts must have attributes that overlap with restriction. 
            Requires restriction with named attributes (e.g. dict, QueryExpression, pandas dataframe)
        :param filter_out_len_zero: (bool) If True, included parts must have greater than zero rows after restriction is applied.
        :param reload_dependencies: (bool) reloads DataJoint graph dependencies.
        """
        assert cls.has_parts(reload_dependencies=reload_dependencies), 'No part tables found. If you are expecting part tables, try with reload_dependencies=True.'

        if include_parts is None:
            parts = cls.parts(as_cls=True)
        
        else:
            parts = cls._format_parts(include_parts)
        
        if exclude_parts is not None:
            parts = [p for p in parts if p.full_table_name not in [e.full_table_name for e in cls._format_parts(exclude_parts)]]
        
        if filter_out_disjoint:
            parts = [p & part_restr for p in parts if not set(p.heading.names).isdisjoint(format_rows_to_df(part_restr).columns)]
        else:
            parts = [p & part_restr for p in parts]

        if filter_out_len_zero:
            parts = [p for p in parts if len(p)>0]

        return parts
    
    @classmethod
    def restrict_one_part(cls, part_restr={}, include_parts=None, exclude_parts=None, filter_out_disjoint=True, filter_out_len_zero=True, reload_dependencies=False):
        """
        Calls `restrict_parts` with filter_out_disjoint=True and filter_out_len_zero=True by default. If not exactly one part table is returned, then a ValidationError will be raised.

        :params: see `restrict_parts`.

        :returns: part table after restriction.
        """
        parts = cls.restrict_parts(part_restr=part_restr, include_parts=include_parts, exclude_parts=exclude_parts, filter_out_disjoint=filter_out_disjoint, filter_out_len_zero=filter_out_len_zero, reload_dependencies=reload_dependencies)

        if len(parts) > 1:
            raise ValidationError('part_restr can restrict multiple part tables.')
        
        elif len(parts) < 1:
            raise ValidationError('part_restr can not restrict any part tables.')
        
        else:
            return unwrap(parts)

    r1p = restrict_one_part # alias for restrict_one_part

    @classmethod
    def part_table_names_with_hash(cls, hash, hash_name=None, include_parts=None, exclude_parts=None, filter_out_len_zero=True, reload_dependencies=False):
        """
        Calls `restrict_parts_with_hash` with filter_out_len_zero=True by default.

        :params: see `restrict_parts_with_hash`

        :returns: list of part table names that contain hash.
        """
        parts = cls.restrict_parts_with_hash(hash=hash, hash_name=hash_name, include_parts=include_parts, exclude_parts=exclude_parts, filter_out_len_zero=filter_out_len_zero, reload_dependencies=reload_dependencies)
        return [format_table_name(r.table_name, part=True) for r in parts]

    @classmethod
    def restrict_one_part_with_hash(cls, hash, hash_name=None, include_parts=None, exclude_parts=None, filter_out_len_zero=True, reload_dependencies=False):
        """
        Calls `restrict_parts_with_hash` with filter_out_len_zero=True by default. If not exactly one part table is returned, then a ValidationError will be raised.

        :params: see `restrict_parts_with_hash`

        :returns: part table after restriction
        """
        parts = cls.restrict_parts_with_hash(hash=hash, hash_name=hash_name, include_parts=include_parts, exclude_parts=exclude_parts, filter_out_len_zero=filter_out_len_zero, reload_dependencies=reload_dependencies)

        if len(parts) > 1:
            raise ValidationError('Hash found in multiple part tables.')
        
        elif len(parts) < 1:
            raise ValidationError('Hash not found in any part tables.')
        
        else:
            return parts[0]
    
    r1pwh = restrict_one_part_with_hash # alias for restrict_one_part_with_hash

    @classmethod
    def restrict_parts_with_hash(cls, hash, hash_name=None, include_parts=None, exclude_parts=None, filter_out_len_zero=False, reload_dependencies=False):
        """
        Checks all part tables and returns the part table that is successfully restricted by {'hash_name': hash}. 

        Note: If hash_name is not provided, cls.hash_name will be tried. 

        A successful restriction is defined by:
            - 'hash_name' is in the part table heading
            - len(part & {'hash_name': hash}) > 0 if filter_out_len_zero=True

        :param hash: hash to restrict with
        :param hash_name: name of attribute that contains hash. If hash_name is None, cls.hash_name will be used.
        :params include_parts, exclude_parts, reload_dependencies: see `restrict_parts`

        :returns: list of part tables after restriction
        """  
        if hash_name is None and hasattr(cls, 'hash_name'):
            hash_name = cls.hash_name

        if hash_name is None:
            raise ValidationError('Table does not have "hash_name" defined, provide it to restrict with hash.')
        
        parts = cls.restrict_parts(part_restr={hash_name: hash}, include_parts=include_parts, exclude_parts=exclude_parts, filter_out_len_zero=filter_out_len_zero, reload_dependencies=reload_dependencies)

        return [p for p in parts if hash_name in p.heading.names]
    
    @classmethod
    def hashes_not_in_parts(cls, hash_name=None, part_restr={}, include_parts=None, exclude_parts=None, filter_out_len_zero=False, reload_dependencies=False):
        """
        Restricts master table to any hashes not found in any of its part tables.

        :param hash_name: name of attribute that contains hash. If hash_name is None, cls.hash_name will be used.
        :params part_restr, include_parts, exclude_parts, reload_dependencies: see `restrict_parts`

        :returns: cls after restriction
        """
        if hash_name is None and hasattr(cls, 'hash_name'):
            hash_name = cls.hash_name

        if hash_name is None:
            raise ValidationError('Table does not have "hash_name" defined, provide it to restrict with hash.')

        return cls - np.sum([(dj.U(cls.hash_name) & p) for p in cls.restrict_parts(part_restr=part_restr, include_parts=include_parts, exclude_parts=exclude_parts, filter_out_len_zero=filter_out_len_zero, reload_dependencies=reload_dependencies)])

    @classmethod
    def insert(cls, rows, replace=False, skip_duplicates=False, ignore_extra_fields=False, allow_direct_insert=None, reload_dependencies=False, insert_to_parts=None, insert_to_parts_kws={}, skip_hashing=False, constant_attrs={}, overwrite_rows=False):
        """
        Insert rows to cls.

        :params rows, replace, skip_duplicates, ignore_extra_fields, allow_direct_insert: see DataJoint insert function.
        :param reload_dependencies (bool): force reload DataJoint networkx graph dependencies before insert.
        :param insert_to_parts: (part table or list of part tables) part table(s) to insert to after master table insert.
        :param insert_to_parts_kws (dict): kwargs to pass to part table insert function.
        :param skip_hashing (bool): If True, hashing will be skipped if hashing is enabled. 
        :param constant_attrs (dict): Python dictionary to add to every row of rows
        :overwrite_rows (bool): Whether to overwrite key/ values in rows. If False, conflicting keys will raise a ValidationError.
        """
        error_msg = 'Error inserting to %s'

        cls.load_dependencies(force=reload_dependencies)

        if not cls._is_insert_validated:
            cls._insert_validation()
        
        rows = cls._prepare_insert(rows, constant_attrs=constant_attrs, overwrite_rows=overwrite_rows, skip_hashing=skip_hashing)
        
        if insert_to_parts is not None:
            if not dj.conn().in_transaction:
                with dj.conn().transaction:
                    try:
                        super().insert(cls(), rows=rows, replace=replace, skip_duplicates=skip_duplicates, ignore_extra_fields=ignore_extra_fields, allow_direct_insert=allow_direct_insert)
                    except:
                        logger.error(error_msg, cls.class_name)
                        raise
                    
                    try:
                        assert cls.has_parts(), 'No part tables found. If you are expecting part tables, try with reload_dependencies=True.'
                        insert_to_parts = cls._format_parts(insert_to_parts)
                        for part in insert_to_parts:
                            part.insert(rows=rows, **{'ignore_extra_fields': True}) if insert_to_parts_kws == {} else part.insert(rows=rows, **insert_to_parts_kws)

                    except:
                        logger.error(error_msg, 'part table.')
                        raise
            else:
                try:
                    super().insert(cls(), rows=rows, replace=replace, skip_duplicates=skip_duplicates, ignore_extra_fields=ignore_extra_fields, allow_direct_insert=allow_direct_insert)
                except:
                    logger.error(error_msg, cls.class_name)
                    raise
                
                try:
                    assert cls.has_parts(), 'No part tables found. If you are expecting part tables, try with reload_dependencies=True.'
                    insert_to_parts = cls._format_parts(insert_to_parts)
                    for part in insert_to_parts:
                        part.insert(rows=rows, **{'ignore_extra_fields': True}) if insert_to_parts_kws == {} else part.insert(rows=rows, **insert_to_parts_kws)

                except:
                    logger.error(error_msg, 'part table.')
                    raise
                    
        else:
            try:
                super().insert(cls(), rows=rows, replace=replace, skip_duplicates=skip_duplicates, ignore_extra_fields=ignore_extra_fields, allow_direct_insert=allow_direct_insert)
            except:
                logger.error(error_msg, cls.class_name)
                raise

    def put(self, replace=False, skip_duplicates=False, ignore_extra_fields=True, allow_direct_insert=None, reload_dependencies=False, insert_to_parts=None, insert_to_parts_kws={}, skip_hashing=False, constant_attrs={}, overwrite_rows=False, **kwargs):
        self.insert1(kwargs, replace=replace, skip_duplicates=skip_duplicates, ignore_extra_fields=ignore_extra_fields, allow_direct_insert=allow_direct_insert, reload_dependencies=reload_dependencies, insert_to_parts=insert_to_parts, insert_to_parts_kws=insert_to_parts_kws, skip_hashing=skip_hashing, constant_attrs=constant_attrs, overwrite_rows=overwrite_rows)

        
class BasePart(Base):
    _is_hash_name_validated = False

    @classmethod
    def _init_validation(cls, **kwargs):
        """
        Validation for initialization of subclasses of abstract class BasePart. 
        """
        super()._init_validation(hash_table_name=cls.hash_table_name)
    
    @classmethod
    def _hash_name_validation(cls, source='self'):
        """
        Validates hash_name and sets hash_len
        """

        part_hash_len = None
        if cls.hash_name in cls.heading.names:
            part_hash_len = _validate_hash_name_type_and_parse_hash_len(cls.hash_name, cls.heading.attributes)

        master_hash_len = None
        if cls.hash_name in cls.master.heading.names:
            master_hash_len = _validate_hash_name_type_and_parse_hash_len(cls.hash_name, cls.master.heading.attributes)

        if part_hash_len and master_hash_len:
            assert part_hash_len == master_hash_len, f'hash_name "{cls.hash_name}" varchar length mismatch. Part table length is {part_hash_len} but master length is {master_hash_len}.'        
            cls._hash_len = part_hash_len
        
        elif part_hash_len:
            cls._hash_len = part_hash_len
        
        else: 
            cls._hash_len = master_hash_len

        cls._is_hash_name_validated = True
    
    @classproperty
    def class_name_valid_id(cls):
        return cls.class_name.replace('.', 'xx')

    @classmethod
    def _insert_validation(cls):
        """
        Validation for insertion into subclasses of abstract class BasePart. 
        """
                    
        if cls.hash_name is not None:
            if not (cls.hash_name in cls.heading.names or cls.hash_name in cls.master.heading.names):
                raise ValidationError(f'hash_name: "{cls.hash_name}" must be present in the part table or master table heading.')
            
            # hash_name validation
            if not cls._is_hash_name_validated:
                cls._hash_name_validation()

        super()._insert_validation()

    @classmethod
    def insert(cls, rows, replace=False, skip_duplicates=False, ignore_extra_fields=False, allow_direct_insert=None, reload_dependencies=False, insert_to_master=False, insert_to_master_kws={}, skip_hashing=False, constant_attrs={}, overwrite_rows=False):
        """
        Insert rows to cls.

        :params rows, replace, skip_duplicates, ignore_extra_fields, allow_direct_insert: see DataJoint insert function
        :param reload_dependencies (bool): force reload DataJoint networkx graph dependencies before insert.
        :param insert_to_master (bool): whether to insert to master table before inserting to part.
        :param insert_to_master_kws (dict): kwargs to pass to master table insert function.
        :param skip_hashing (bool): If True, hashing will be skipped if hashing is enabled. 
        :param constant_attrs (dict): Python dictionary to add to every row in rows
        :overwrite_rows (bool): Whether to overwrite key/ values in rows. If False, conflicting keys will raise a ValidationError.
        """
        error_msg = 'Error inserting to %s'

        assert isinstance(insert_to_master, bool), '"insert_to_master" must be a boolean.'
        
        cls.load_dependencies(force=reload_dependencies)
        
        rows = cls._prepare_insert(rows, constant_attrs=constant_attrs, overwrite_rows=overwrite_rows, skip_hashing=skip_hashing)

        if insert_to_master:
            if not dj.conn().in_transaction:
                with dj.conn().transaction:
                    cls.master.insert(rows=rows, **{'ignore_extra_fields': True, 'skip_duplicates': True}) if insert_to_master_kws == {} else cls.master.insert(rows=rows, **insert_to_master_kws)

                    try:
                        super().insert(cls(), rows=rows, replace=replace, skip_duplicates=skip_duplicates, ignore_extra_fields=ignore_extra_fields, allow_direct_insert=allow_direct_insert)
                    except:
                        logger.error(error_msg, cls.class_name)
                        raise
            else:
                cls.master.insert(rows=rows, **{'ignore_extra_fields': True, 'skip_duplicates': True}) if insert_to_master_kws == {} else cls.master.insert(rows=rows, **insert_to_master_kws)

                try:
                    super().insert(cls(), rows=rows, replace=replace, skip_duplicates=skip_duplicates, ignore_extra_fields=ignore_extra_fields, allow_direct_insert=allow_direct_insert)
                except:
                    logger.error(error_msg, cls.class_name)
                    raise

        else:
            try:
                super().insert(cls(), rows=rows, replace=replace, skip_duplicates=skip_duplicates, ignore_extra_fields=ignore_extra_fields, allow_direct_insert=allow_direct_insert)
            except:
                logger.error(error_msg, cls.class_name)
                raise
    
    def put(self, replace=False, skip_duplicates=False, ignore_extra_fields=True, allow_direct_insert=None, reload_dependencies=False, insert_to_master=False, insert_to_master_kws={}, skip_hashing=False, constant_attrs={}, overwrite_rows=False, **kwargs):
        self.insert1(kwargs, replace=replace, skip_duplicates=skip_duplicates, ignore_extra_fields=ignore_extra_fields, allow_direct_insert=allow_direct_insert, reload_dependencies=reload_dependencies, insert_to_master=insert_to_master, insert_to_master_kws=insert_to_master_kws, skip_hashing=skip_hashing, constant_attrs=constant_attrs, overwrite_rows=overwrite_rows)



"""
DataJointPlus is an extension to DataJoint. DataJointPlus extends DataJoint's 
user tables with automatic hashing and enhances the DataJoint 
master-part relationship, providing templates for commonly used motifs.
"""

__all__ = ['__version__',
           'kill', 'set_password', 'AndList', 
           'Not', 'U', 'Connection', 'conn', 
           'AttributeAdapter', 'MatCell', 'MatStruct', 
           'Diagram', 'DataJointError', 'key',
           'list_schemas', 'config',
           'DataJointPlusModule', 'create_djp_module',
           'reassign_master_attribute', 'add_datajoint_plus',
           'Lookup', 'Computed', 'Part',
           'add_objects', 'check_if_latest_version', 'enable_datajoint_flags',
           'format_table_name', 'split_full_table_name', 'make_store_dict', 'register_externals', 
           'generate_hash', 'validate_and_generate_hash', 'parse_definition', 
           'reform_definition', 'errors', 'free_table', 
           'basicConfig', 'getLogger', 'LogFileManager']



# from DataJoint
from datajoint.admin import kill, set_password
from datajoint.attribute_adapter import AttributeAdapter
from datajoint.blob import MatCell, MatStruct
from datajoint.connection import Connection, conn
from datajoint.diagram import Diagram
from datajoint.errors import DataJointError
from datajoint.expression import AndList, Not, U
from datajoint.fetch import key
from datajoint.schemas import VirtualModule as DataJointVirtualModule, list_schemas 

# from DataJointPlus
from . import errors
from .compatibility import add_datajoint_plus, reassign_master_attribute
from .config import config
from .hash import generate_hash, validate_and_generate_hash
from .heading import parse_definition, reform_definition
from .logging import LogFileManager, basicConfig, getLogger
from .schema import DataJointPlusModule, Schema
from .table import FreeTable as free_table
from .user_tables import Computed, Lookup, Part
from .utils import (add_objects, check_if_latest_version,
                    enable_datajoint_flags, format_table_name, make_store_dict,
                    register_externals, split_full_table_name)
from .version import __version__

# aliases
ERD = Di = Diagram                      
create_dj_virtual_module = DataJointVirtualModule
schema = Schema 
create_djp_module = DataJointPlusModule 

# version control
check_if_latest_version()




"""
User tables for DataJointPlus Motifs
"""
from collections import namedtuple
import inspect
import logging
import re
import datajoint as dj
import numpy as np
from datajoint_plus.definition import StrToTable
from datajoint_plus.errors import MakerDestinationError, MakerError, MakerInputError, MakerMethodError
from datajoint.fetch import is_key
from datajoint.expression import Projection
from datajoint_plus.user_tables import UserTable

from .base import BaseMaster, BasePart
from .utils import classproperty, safedict, unwrap, wrap

logger = logging.getLogger(__name__)






class Entity(StrToTable):
    is_entity = True
    def __init__(self, source, add_to_key_source:bool=True, inheritance:bool='primary', get=None, context=None, **kwargs):
        if getattr(source, 'is_entity', False):
            self.__dict__ = source.__dict__
        else:
            if is_key(get):
                self.get = self.get_key
            super().__init__(source=source, add_to_key_source=add_to_key_source, inheritance=inheritance, get=get, context=context, **kwargs)
    
    def get_key(self, key):
        return (self.table & key).fetch1('KEY')
    
    
class Method(StrToTable):
    is_method = True
    def __init__(self, source, add_to_key_source:bool=True, inheritance:bool='primary', run=None, context=None, **kwargs):
        if getattr(source, 'is_method', False):
            self.__dict__ = source.__dict__
        else:
            super().__init__(source=source, add_to_key_source=add_to_key_source, inheritance=inheritance, run=run, context=context, **kwargs)

        
class Destination(StrToTable):
    is_destination = True
    def __init__(self, source, inheritance:bool='primary', put=None, context=None, **kwargs):
        if getattr(source, 'is_destination', False):
            self.__dict__ = source.__dict__
        else:
            super().__init__(source=source, inheritance=inheritance, put=put, context=context, **kwargs)


class Motif:
    """
    Class level: -2
    """

    @classmethod
    def _init_validation(cls, **kwargs):
        if (cls.hash_name is None) and (cls.lookup_name is None):
            raise NotImplementedError('Subclasses of Motif must implement "lookup_name" or "hash_name".')
    
    @staticmethod
    def _emd_mapping(emd_type:str):
        prefix_mapping = {
            'entities': 'get',
            'methods': 'run',
            'destinations': 'put'
        }

        cls_mapping = {
            'entities': Entity,
            'methods': Method,
            'destinations': Destination,
        }
        try:
            return prefix_mapping[emd_type], cls_mapping[emd_type]
        except KeyError:
            msg = f'emd_type {emd_type} not recognized.'
            logger.error(msg)
            raise AttributeError(msg)
    
    @classmethod
    def _emd_namedtuple(cls, emd_type:str, context=None):
        emd_prefix, emd_cls = cls._emd_mapping(emd_type)
        emd = getattr(cls, emd_prefix + '_' + emd_type, None)
        if emd is not None:
            emds = []
            for source in wrap(emd):
                emds.append(
                    emd_cls(
                        source, 
                        context=context
                    )
                )
        if emds is not None:
            nt = namedtuple(
                emd_type, 
                field_names=[getattr(t.table, 'class_name_valid_id', getattr(t.table, 'class_name')) for t in emds]
            )
            nt.__repr__ = cls.nt_repr
            return nt(*emds)
        else:
            return namedtuple(emd_type, field_names='')()
    
    @classmethod
    def str_to_table(cls, source:str, context=None):
        return StrToTable(source, context=context).table

    @classmethod
    def str_to_base(cls, source:str, context=None):
        return StrToTable(source, context=context).base


class EntityLookup(Motif, BaseMaster, UserTable, dj.Lookup):
    @classmethod
    def _init_validation(cls, **kwargs):
        super(Motif, cls)._init_validation(**kwargs)
        super()._init_validation(**kwargs)
    
    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)
    
    def get(self, key={}, attrs={}):
        if self.stores is None:
            return super().get(key=key, attrs=attrs)
        else:
            data = []        
            for store in wrap(self.stores):
                store = eval(store)
                if store.full_table_name in self.descendants():
                    key = (self & key).proj(**{store.lookup_name: self.lookup_name}).fetch('KEY')
                else:
                    raise
                d = store().get(key=key, attrs=attrs)
                data.append(d)
            return unwrap(data)


class MethodLookup(Motif, BaseMaster, UserTable, dj.Lookup):
    @classmethod
    def _init_validation(cls, **kwargs):
        super(Motif, cls)._init_validation(**kwargs)
        super()._init_validation(**kwargs)
    
    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)

    def run(self, **kwargs):
        return self.r1p(kwargs).run(**kwargs)


class StoreLookup(Motif, BaseMaster, UserTable, dj.Lookup):    
    @classmethod
    def _init_validation(cls, **kwargs):
        super(Motif, cls)._init_validation(**kwargs)
        super()._init_validation(**kwargs)
    
    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)
        
    def get(self, key, **kwargs):
        return self.r1p(key).get(**kwargs)
    
    def put(self, parts, **kwargs):
        for p in parts:
            p.put(**kwargs)


class MakerLookup(Motif, BaseMaster, UserTable, dj.Lookup):
    @classmethod
    def _init_validation(cls, **kwargs):
        super(Motif, cls)._init_validation(**kwargs)
        super()._init_validation(**kwargs)
    
    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)
    
    @classmethod
    def populate(cls, *args, **kwargs):
        for p in cls.parts(as_cls=True):
            p.populate(*args, **kwargs)


class NestedMethod(Motif, BasePart, UserTable, dj.Part):
    @classmethod
    def _init_validation(cls, **kwargs):
        super(Motif, cls)._init_validation(**kwargs)
        super()._init_validation(**kwargs)
        
    def __init_subclass__(cls, **kwargs):
        cls.enable_hashing = True
        cls._init_validation(**kwargs)


class NestedStore(Motif, BasePart, UserTable, dj.Part):
    @classmethod
    def _init_validation(cls, **kwargs):
        super(Motif, cls)._init_validation(**kwargs)
        super()._init_validation(**kwargs)
    
    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)


class NestedMaker(Motif, BasePart, UserTable, dj.Part, dj.Computed):
    definition = None
    get_entities = None
    run_methods = None
    put_destinations = None
    _dict_merge_warn_overwrite = True
    _dict_merge_allow_overwrite = False
    _append_timestamp_to_definition = True
    _timestamp_name = 'ts_inserted'
    
    def _init_validation(cls, **kwargs):
        super(Motif, cls)._init_validation(**kwargs)
        super()._init_validation(**kwargs)

    def __init_subclass__(cls, **kwargs):
        # HASHED ATTRS
        if getattr(cls, 'hashed_attrs', None) == 'key_source':
            cls.hashed_attrs = cls.key_source.primary_key
           
        # DEFINITION
        # make definition
        if cls.definition is None:
            # sort dependencies
            foreign_key = {'primary': [], 'secondary': []}
            for ps in ['primary', 'secondary']:
                for emd_type in ['entities', 'methods', 'destinations']:
                    emd_nt = getattr(cls, emd_type)
                    for emd in emd_nt:
                        if emd.inheritance == ps:
                            foreign_key[ps].extend(f"-> self.str_to_base(**{{'emd_type': '{emd_type}', 'source': '{emd.source}', 'context': self.declaration_context}}) \n") 
            
            definition = ''.join([
                "-> master \n",
                "".join(foreign_key['primary']),
                """
                ---
                """,
                "".join(foreign_key['secondary'])
            ])
            if cls._append_timestamp_to_definition:
                definition += f"""
                {cls._timestamp_name}=CURRENT_TIMESTAMP: timestamp # 
                """
            
            # format definition
            f_definition = []
            for line in re.split("\n", definition):
                line = line.strip(' ')
                if line:
                    f_definition.append(line)
            cls.definition = '\n '.join(f_definition)
            
        cls._init_validation(cls, **kwargs)
    
    @classproperty
    def entities(cls):
        return cls._emd_namedtuple('entities', context=cls.declaration_context)
    
    @classproperty
    def methods(cls):
        return cls._emd_namedtuple('methods', context=cls.declaration_context)
    
    @classproperty
    def destinations(cls):
        return cls._emd_namedtuple('destinations', context=cls.declaration_context)
    
    @classproperty
    def upstream(cls):
        return namedtuple('upstream', ['entities', 'methods'])(cls.entities, cls.methods)
    
    @classproperty
    def downstream(cls):
        return namedtuple('downstream', ['destinations'])(cls.destinations)
    
    @classproperty
    def key_source(cls):
        ks = []
        for emd_type in ['get_entities', 'run_methods']:
            if getattr(cls, emd_type) is not None:
                items = getattr(cls, emd_type.split('_')[1])
                for i in items:
                    if i.add_to_key_source:
                        ks.append(i.table)
        ks = np.product(ks)
        assert isinstance(ks, dj.Table), 'key_source must be a DataJoint table. Check that get_entities and/ or run_methods was defined correctly.'
        return ks
    
    def _extract_fxn(self, emd):
        """
        Extracts the "get", "run" or "put" function from the input.
        """
        if getattr(emd, 'is_entity', False):
            fxn_type = 'get'
            err = MakerInputError
        elif getattr(emd, 'is_method', False):
            fxn_type = 'run'
            err = MakerMethodError
        elif getattr(emd, 'is_destination', False):
            fxn_type = 'put'
            err = MakerDestinationError
        else:
            raise AttributeError(f'type {type(emd)} not recognized. Expected djp.Entity, djp.Method or djp.Destination.')
        
        fxn = getattr(emd, fxn_type, None) or getattr(emd.table, fxn_type, None)

        if fxn is None:
            msg = f'"{fxn_type}" function not found for {emd}.'
            logger.error(msg)
            raise err(msg)
        return fxn
    
    @staticmethod
    def _validate_arg(arg):
        """
        Validates and returns arg, or errors if validation fails.
        """
        if not isinstance(arg, dict):
            msg = f'arg should be dict instance.'
            logger.error(msg)
            raise MakerError(msg)
        return arg
    
    def make(self, key):
        # GET ENTITIES
        inputs = safedict(warn=self._dict_merge_warn_overwrite, overwrite=self._dict_merge_allow_overwrite)
        if self.entities:
            for e in self.entities:
                get = self._extract_fxn(e)
                inp = self._validate_arg(get(key))
                inputs.update(**inp)
        inputs = {**key, **inputs}

        # RUN METHODS
        results = safedict(warn=self._dict_merge_warn_overwrite, overwrite=self._dict_merge_allow_overwrite)
        if self.methods:
            for m in self.methods:
                run = self._extract_fxn(m)
                res = self._validate_arg(run(**inputs))
                results.update(**res)
        results = {**key, **results}
        
        # HASH
        hash_dict = {self.hash_name: self.hash1(results)} if self.enable_hashing else {}
        row = {**hash_dict, **results}

        # PUT DESTINATIONS
        if self.destinations:
            for d in self.destinations:
                put = self._extract_fxn(d)
                put(**row)

        self.put(**row, skip_hashing=True, insert_to_master=True, insert_to_master_kws={'ignore_extra_fields': True, 'skip_duplicates': True})
    
    def nt_repr(self):
        base = self.__class__.__name__ + '('
        for i, f in enumerate(self._fields):
            try:
                self._fields[i+1]
            except IndexError:
                base += f + ')'
            else:
                base += f + ', '
        return base


"""
Tools to validate user input.
"""

import re
from itertools import combinations

import numpy as np

from .errors import OverwriteError


def pairwise_disjoint_set_validation(sets:list, set_names:list=None, error=Exception):
    """
    Checks all pairs of sets in provided list for disjointness. Will raise error if any two sets are not disjoint.
    :param sets: list of sets to check.
    :param set_names: (optional) list of set names to index and provide in error message if disjoint check fails. \
        Length and order must match "sets". Defaults to generic error message.
    :param error: (optional) error to throw upon validation failure. Defaults to Exception.
    :returns: None if validation passes, error if fail. 
    """
    if set_names is not None:
        assert len(sets) == len(set_names), 'Length of sets must match length of set_names'

    set_combinations = list(combinations(np.arange(len(sets)), 2))
    for c in set_combinations:
        if not set.isdisjoint(sets[c[0]], sets[c[1]]):
            if set_names is not None:
                raise error(f'attributes in "{set_names[c[0]]}" and "{set_names[c[1]]}" must be disjoint.')
            else:
                raise error(f'attributes in at least two provided sets are not disjoint.')


def _validate_hash_name_type_and_parse_hash_len(hash_name, attributes):
    """
    Validates the attribute type of hash_name and extracts the character length of hash.

    :param hash_name: (str) hash_name to validate.
    :param attributes: (dict) dj_table.heading.attributes dictionary that hash_name will index into.

    :returns: 
        - error if validation fails
        - hash character length (int) if validation passes
    """
    try:
        hash_type = attributes[hash_name].type
    except KeyError:
        raise KeyError(f'hash_name "{hash_name}" not found in attributes.') from None

    _, m, e = hash_type.rpartition('varchar')
    assert m == 'varchar', 'hash_name attribute must be of varchar type'

    hash_len_match = re.findall('[0-9]+', e)
    assert hash_len_match and hash_len_match[0].isdigit(), 'hash_name attribute must contain a numeric value specifying hash character length.'
    
    hash_len = int(hash_len_match[0])
    assert hash_len > 0 and hash_len <= 32, 'hash character length must be within range: [1, 32].'

    return hash_len


def _is_overwrite_validated(attr, group, overwrite_rows):
    """
    Checks if attr is in group and is overwriteable.
    """
    assert isinstance(overwrite_rows, bool), 'overwrite_rows must be a boolean.'
    if not overwrite_rows:
        if attr in group:
            raise OverwriteError(f'Attribute "{attr}" already in rows. To overwrite, set overwrite_rows=True.')
        else:
            return True
    return True


import os
from enum import Enum

from datajoint.settings import config

class DEFAULTS(Enum):
    DJ_LOGLEVEL = 'INFO'
    DJ_LOG_BASE_DIR = 'logs' 

for default in DEFAULTS:
    if os.getenv(default.name) is None:
        os.environ[default.name] = default.value

config_mapping = {
    'loglevel': DEFAULTS.DJ_LOGLEVEL,
    'log_base_dir': DEFAULTS.DJ_LOG_BASE_DIR,
}

for k, v in config_mapping.items():
    config.update({k: os.getenv(v.name)})




"""General-purpose utilities"""

import inspect
import logging
import re
import sys
from unittest import mock

import numpy as np
import pandas as pd
import requests
from datajoint.errors import _switch_adapted_types, _switch_filepath_types
from datajoint.table import QueryExpression
from datajoint.user_tables import UserTable
from IPython.display import display
from ipywidgets.widgets import HBox, Label, Output

from .config import config
from .errors import OverwriteError, ValidationError
from .hash import generate_table_id
from .logging import getLogger
from .version import __version__

logger = getLogger(__name__)

class classproperty:
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


def wrap(item):
    if not isinstance(item, list) and not isinstance(item, tuple):
        item = [item]
    return item


def unwrap(item):
    if isinstance(item, list) or isinstance(item, tuple):
        if len(item) == 1:
            return item[0]
    return item


def split_full_table_name(full_table_name:str):
    """
    Splits full_table_name from DataJoint tables and returns a tuple of (database, table_name).

    :param (str): full_table_name from DataJoint tables
    
    :returns (tuple): (database, table_name)
    """
    return tuple(s.strip('`') for s in full_table_name.split('.'))


def reform_full_table_name(schema_name:str, table_name:str):
    """
    Reforms full_table_name from DataJoint schema name and a table_name.

    :param schema_name (str): name of schema
    :param table_name (str): name of table
    
    :returns: full_table_name
    """
    return '.'.join(['`'+schema_name+'`', '`'+table_name+'`'])


def format_table_name(table_name, snake_case=False, part=False):
    """
    Splits full_table_name from DataJoint tables and returns a tuple of (database, table_name).

    :param (str): full_table_name from DataJoint tables
    
    :returns (tuple): (database, table_name)
    """
    if not snake_case:
        if not part:
            return table_name.title().replace('_','').replace('#','')
        else:
            return table_name.title().replace('__','.').replace('_','').replace('#','')
    else:
        if not part:
            return table_name.lower().strip('_').replace('#','')
        else:
            return table_name.lower().replace('__','.').strip('_').replace('#','')


def format_rows_to_df(rows):
    """
    Formats rows as pandas dataframe.
    :param rows: pandas dataframe, datajoint query expression, dict or tuple
    :returns: pandas dataframe
    """
    if isinstance(rows, pd.DataFrame):
        rows = rows.copy()
    elif (inspect.isclass(rows) and issubclass(rows, QueryExpression)) or isinstance(rows, QueryExpression):
        rows = pd.DataFrame(rows.fetch())
    elif isinstance(rows, list) or isinstance(rows, tuple):
        rows = pd.DataFrame(rows)
    elif isinstance(rows, dict):
        rows = pd.DataFrame([rows])
    else:
        raise ValidationError('Format of rows not recognized. Try a list of dictionaries, a DataJoint expression or a pandas dataframe.')

    return rows


def load_dependencies(connection, force=False, verbose=True):
    """
    Loads dependencies in a DataJoint connection object.

    :param connection: (datajoint.connection) DataJoint connection object 
    :param force: (bool) default False. Whether to force reload.
    """
    if verbose:
        if force:
            output = Output()
            display(output)
            with output:
                pending_text = Label('Loading schema dependencies...')
                confirmation = Label('Success.')
                confirmation.layout.display = 'none'
                display(HBox([pending_text, confirmation]))
                connection.dependencies.load()
                confirmation.layout.display = None
    else:
        connection.dependencies.load(force=force)


def enable_datajoint_flags(enable_python_native_blobs=True):
    """
    Enable experimental datajoint features
    
    These flags are required by 0.12.0+ (for now).
    """
    config['enable_python_native_blobs'] = enable_python_native_blobs
    _switch_filepath_types(True)
    _switch_adapted_types(True)


def register_externals(external_stores):
    """
    Registers external stores to DataJoint.
    """
    if 'stores' not in config:
        config['stores'] = external_stores
    else:
        config['stores'].update(external_stores)


def make_store_dict(path):
    return {
        'protocol': 'file',
        'location': str(path),
        'stage': str(path)
    }


def _get_calling_context() -> locals:
    # get the calling namespace
    try:
        frame = inspect.currentframe().f_back
        context = frame.f_locals
    finally:
        del frame
    return context


def add_objects(objects, context=None):
    """
    Imports the adapters for a schema_name into the global namespace.
    """   
    if context is None:
        # if context is missing, use the calling namespace
        try:
            frame = inspect.currentframe().f_back
            context = frame.f_locals
        finally:
            del frame
    
    for name, obj in objects.items():
        context[name] = obj


def check_if_latest_version(source='github', return_latest=False):
    """
    Checks if imported DataJointPlus version matches latest from source. Logs warning if versions do not match.

    :param source: (str) Options:
        github
    :param return_latest: (bool) If True, returns the latest version
    """
    try:
        if source == 'github':
            _latest_version_text = re.search('__version__.*', requests.get(f"https://raw.githubusercontent.com/cajal/datajoint-plus/main/datajoint_plus/version.py").text).group()
            latest_version = _latest_version_text.split('=')[1].strip(' "'" '") if len(_latest_version_text.split('='))>1 else _latest_version_text.strip(' "'" '")
            if __version__ != latest_version:
                logger.warning(f'Imported datajoint_plus version, {__version__} does not match the latest version on Github, {latest_version}.')
        else:
            raise AttributeError('Source not recognized. "github" is the only supported source')

        if return_latest:
            return latest_version
    except:
        logger.warning(f'DataJointPlus version check failed.')


def goto(table_id=None, full_table_name=None, directory='__main__', warn=True):
    """
    Checks table_id's of DataJoint user classes in the current module and returns the class if a partial match to table_id or full_table_name is found. 
    
    :param table_id: (str) table_id to check (found in user_class.table_id and schema.tables)
    :param full_table_name: (str) full_table_name to check (found in user_class.full_table_name and schema.tables)
    :param directory: Options - 
        '__main__' (str) - default. searches calling module.
        directory - directory to search
    :param warn: (bool) If True, logs warnings
    returns: class if a table_id match is found, otherwise None
    """
    # handle table_id and full_table_name input
    assert ~((table_id is None) and (full_table_name is None)), 'Provide table_id or full_table_name'

    if full_table_name is not None:
        table_id_eval = generate_table_id(full_table_name)
        if table_id is not None:
            assert table_id == table_id_eval, "Provided table_id and table_id evaluated from full_table_name do not match."
        table_id = table_id_eval

    match = []
    def check_directory(d):
        for name, obj in inspect.getmembers(d):
            if name in ['key_source', '_master', 'master', 'UserTable']:
                continue
            if inspect.isclass(obj) and issubclass(obj, UserTable):
                try:
                    if table_id in obj.table_id:
                        match.append(obj)
                        return
                        
                    check_directory(obj)
                except:
                    if warn:
                        logger.warning(f'Could not check table_id for {name}')
                    continue
                
    
    if directory == '__main__':
        directory = sys.modules[directory]
    
    check_directory(directory)
    
    n_unique_matches = len(np.unique([m.table_id for m in match]))
    if n_unique_matches == 1:
        return match[0]
    elif n_unique_matches > 1:
        if warn:
            logger.warning(f'table_id matched to multiple tables.')
    elif n_unique_matches == 0:
        if warn:
            logger.warning(f'table_id did not match to any tables. Are you searching the correct directory?')


def user_choice_with_default_response(default_response=None):
    """Creates a replacement for the DataJoint `user_choice` function that will
    return a default response if one was provided."""

    def _user_choice(prompt, choices=("yes", "no"), default=None):
        """
        Prompts the user for confirmation.  The default value, if any, is capitalized.
        :param prompt: Information to display to the user.
        :param choices: an iterable of possible choices.
        :param default: default choice
        :param default_response: If default_response is provided, bypasses input and returns default_response
        :return: the user's choice
        """
        if default_response is not None:
            return default_response
        
        assert default is None or default in choices
        choice_list = ', '.join((choice.title() if choice == default else choice for choice in choices))
        response = None
        while response not in choices:
            response = input(prompt + ' [' + choice_list + ']: ')
            response = response.lower() if response else default
        return response

    return _user_choice


def default_user_choice(default:str):
    return mock.patch("datajoint.table.user_choice", new=user_choice_with_default_response(default))


class safedict(dict):
    err_msg = 'Cannot update safedict because overwrite = False'

    def __init__(self, warn=True, overwrite=False, logger=None, *args, **kwargs):
        """
        Extension of dict that can prevent update if key already present in dict.

        :param warn:  (bool) toggle warnings
        :param overwrite: (bool) 
            if False, dict update will be prevent if key found in dict. 
            if True, defaults to normal dict behavior
        :param logger: logger to use for warnings. defaults to generic logging. 
        """
        self.warn = warn
        self.overwrite = overwrite
        self.logger = logger if logger is not None else logging.getLogger('__main__')
        super().__init__(*args, **kwargs)
        
    def _key_in_dict(self, *args, **kwargs):
        in_dict = False
        for keys in [args, kwargs]:
            for key in keys:
                if isinstance(key, dict):
                    key = list(key.keys())[0]
                if key in self:
                    in_dict = True
                    if self.warn:
                        self.logger.warning(f'{key} already in safedict.')
        return in_dict
                    
    def update(self, *args, **kwargs):
        if self._key_in_dict(*args, **kwargs) and not self.overwrite:
            self.logger.error(self.err_msg)
            raise OverwriteError(self.err_msg)
        super().update(*args, **kwargs)
    
    def __setitem__(self, key, value):
        if self._key_in_dict(key) and not self.overwrite:
            self.logger.error(self.err_msg)
            raise OverwriteError(self.err_msg)
        super().__setitem__(key, value)



"""
Motif tables for DataJointPlus
"""

from datajoint_plus.user_tables import UserTable
from datajoint_plus.utils import classproperty
import datajoint as dj


class Motif(UserTable):
    @classmethod
    def is_motif_table(cls):
        return issubclass(Motif)


class MotifMaster(Motif):
    @classproperty
    def definition(cls):
        return f"""
        #{cls().class_name}
        {cls.hash_name} : varchar({cls.hash_len}) # hash
        """


class Nested(Motif):
    @classmethod
    def is_nested(cls):
        return issubclass(cls, Nested)

    def delete_from_master(self):
        # with dj.conn().transaction:
        keys = self.fetch('KEY')
        (self.master & keys).delete()
            
    def delete_quick_from_master(self):
        # with dj.conn().transaction:
        keys = self.fetch('KEY')
        self.delete_quick()
        (self.master & keys).delete_quick()
            
    def delete_quick(self, delete_from_master=True):
            if delete_from_master:
                self.delete_quick_from_master()
            else:
                super().delete_quick()
        
    def delete(self, delete_from_master=True):
        if delete_from_master:
            self.delete_from_master()
        else:
            super().delete()

    def drop(self, force=False):
        if force:
            super(UserTable, self).drop()
        else:
            raise dj.DataJointError('Cannot drop a Part directly.  Delete from master instead')

    def drop_quick(self, force=False):
        if force:
            return super(UserTable, self).drop_quick()
        else:
            raise dj.DataJointError('Cannot drop a Part directly.  Delete from master instead')






"""
Hosts the original DataJoint table tiers extended with DataJointPlus.
"""
from .logging import getLogger
import re

import datajoint as dj

from .table import Table
from .base import BaseMaster, BasePart

master_classes = (dj.Manual, dj.Lookup, dj.Computed, dj.Imported,)
part_classes = (dj.Part,)

logger = getLogger(__name__)

class UserTable(Table, dj.user_tables.UserTable):
    @property
    def is_user_table(self):
        return True

    @classmethod
    def is_master(cls):
        try:
            next(tier for tier in master_classes
                    if re.fullmatch(tier.tier_regexp, cls.table_name))
            return True
        except StopIteration:
            return False

    @classmethod
    def is_part(cls):
        try:
            next(tier for tier in part_classes
                        if re.fullmatch(tier.tier_regexp, cls.table_name))
            return True
        except StopIteration:
            return False


class Lookup(BaseMaster, UserTable, dj.Lookup):
    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)
    

class Manual(BaseMaster, UserTable, dj.Manual):
    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)


class Computed(BaseMaster, UserTable, dj.Computed):
    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)


class Imported(BaseMaster, UserTable, dj.Imported):
    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)


class Part(BasePart, UserTable, dj.Part):
    
    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)

    def drop(self, force=False):
        if force:
            super(UserTable, self).drop()
        else:
            raise dj.DataJointError('Cannot drop a Part directly.  Delete from master instead')

    def drop_quick(self, force=False):
        if force:
            return super(UserTable, self).drop_quick()
        else:
            raise dj.DataJointError('Cannot drop a Part directly.  Delete from master instead')


#!/usr/bin/env python
from setuptools import setup, find_packages
from os import path

here = path.abspath(path.dirname(__file__))

# read in version number into __version__
with open(path.join(here, 'datajoint_plus', 'version.py')) as f:
    exec(f.read())

setup(
    name='datajoint-plus',
    version=__version__,
    description="A DataJoint extension that integrates hashes and other features.",
    author='Stelios Papadopoulos',
    author_email='spapadop@bcm.edu',
    license="GNU LGPL",
    packages=find_packages(),
    install_requires=['datajoint==0.12.9', 'envyaml'],
    python_requires='~=3.6'
)


"""
DataJointPlus Schema extensions
"""

from .logging import getLogger
import types

import datajoint as dj
from .compatibility import add_datajoint_plus
from .utils import enable_datajoint_flags, load_dependencies, register_externals, split_full_table_name, reform_full_table_name
from .table import TableLog
from .hash import generate_table_id
from .utils import classproperty
from .table import FreeTable

logger = getLogger(__name__)

class Schema(dj.Schema):
    """
    Extension of dj.Schema that adds a table log

    Additional params:
    :param load_dependencies (bool): Loads the DataJoint graph.
    """
    def __init__(self, schema_name, context=None, load_dependencies=True, *, connection=None, create_schema=True, create_tables=True):
        super().__init__(schema_name=schema_name, context=context, connection=connection, create_schema=create_schema, create_tables=create_tables)

        # attempt to update ~tables
        try:
            self._tables = None
            for table_name in self.list_tables():
                full_table_name = reform_full_table_name(self.database, table_name)
                self.tables(generate_table_id(full_table_name), full_table_name, action='add')
            for key in self._tables:
                _, name = split_full_table_name(key['full_table_name'])
                if name not in self.list_tables():
                    self.tables(full_table_name=key['full_table_name'], action='delete')
        except:
            pass

        self.load_dependencies(force=load_dependencies, verbose=False)


    @classproperty
    def is_schema(cls):
        True

    @property
    def tables(self):
        if self._tables is None:
            self._tables = TableLog(self.connection, self.database)
        return self._tables

    def free_table(self, table_name=None, full_table_name=None):
        """
        Generates a free table in the schema from table_name or full_table_name.
        
        :param table_name: (str) Name of table in MySQL. 
        :param full_table_name: (str) Full table name with format '`database`.`table_name`'

        :returns: dj.FreeTable
        """
        table_not_in_schema = 'Table not found in schema. See tables with schema.list_tables().'
        assert (table_name is None) ^ (full_table_name is None), 'Provide table_name or full_table_name but not both'
        if table_name is not None:
            assert table_name in self.list_tables(), table_not_in_schema
            full_table_name = reform_full_table_name(self.database, table_name)
        elif full_table_name is not None:
            try:
                database, _ = split_full_table_name(full_table_name)
            except ValueError:
                raise ValueError('Split full_table_name failed. Did you pass a table_name in full_table_name arg?')
            assert database == self.database, table_not_in_schema
        else:
            raise AttributeError('Provide table_name or full_table_name.')
        return FreeTable(self.connection, full_table_name)
    
    def load_dependencies(self, force=True, verbose=True):
        """
        Loads dependencies into DataJoint networkx graph. 
        """
        load_dependencies(self.connection, force=force, verbose=verbose)


class VirtualModule(types.ModuleType):
    """
    A virtual module which will contain context for schema.
    """
    def __init__(self, module_name, schema_name, *, create_schema=False,
                 create_tables=False, connection=None, add_objects=None):
        """
        Creates a python module with the given name from the name of a schema on the server and
        automatically adds classes to it corresponding to the tables in the schema.
        :param module_name: displayed module name
        :param schema_name: name of the database in mysql
        :param create_schema: if True, create the schema on the database server
        :param create_tables: if True, module.schema can be used as the decorator for declaring new
        :param connection: a dj.Connection object to pass into the schema
        :param add_objects: additional objects to add to the module
        :return: the python module containing classes from the schema object and the table classes
        """
        super(VirtualModule, self).__init__(name=module_name)
        _schema = Schema(schema_name, create_schema=create_schema, create_tables=create_tables,
                         connection=connection)
        if add_objects:
            self.__dict__.update(add_objects)
        self.__dict__['schema'] = _schema
        _schema.spawn_missing_classes(context=self.__dict__)


class DataJointPlusModule(VirtualModule):
    """
    DataJointPlus extension of DataJoint virtual module with the added ability to instantiate from an existing module.
    """
    def __init__(self, module_name=None, schema_name=None, module=None, schema_obj_name=None, add_externals=None, add_objects=None, create_schema=False, create_tables=False, connection=None, spawn_missing_classes=True, load_dependencies=True, enable_dj_flags=True, warn=True):
        """
        Add DataJointPlus methods to all DataJoint user tables in a DataJoint virtual module or to an existing module. 
        
        To instantiate a DataJoint Virtual Module, provide args module_name and schema_name. 
        
        To modify an existing module, provide arg module. 
        
        :param module_name (str): displayed module name (if using DataJoint Virtual module)
        :param schema_name (str): name of the database in mysql
        :param module (module): module to modify
        :param schema_obj_name (str): The name of the schema object you wish to instantiate (only needed if the module contains more than one DataJoint dj.schema object)
        :param add_externals (dict): Dictionary mapping to external files.
        :param add_objects (dict): additional objects to add to the module
        :param spawn_missing_classes (bool): Only relevant if module provided. If True, adds DataJoint tables not in module but present in mysql as classes. 
        :param load_dependencies (bool): Loads the DataJoint graph.
        :param create_schema (bool): if True, create the schema on the database server
        :param create_tables (bool): if True, module.schema can be used as the decorator for declaring new
        :param connection (dj.Connection): a dj.Connection object to pass into the schema
        :param enable_dj_flags (bool): If true runs djp.enable_datajoint_flags. May be necessary to use adapters. 
        :param warn (bool): if False, warnings are disabled. 
        :return: the virtual module or modified module with DataJointPlus added.
        """
        if schema_name:
            assert not module, 'Provide either schema_name or module but not both.'
            super().__init__(module_name=module_name if module_name else schema_name, schema_name=schema_name, add_objects=add_objects, create_schema=create_schema, create_tables=create_tables, connection=connection)
            
            self.load_dependencies(self.__dict__['schema'].connection, load_dependencies, verbose=False)
            
        elif module:
            super(dj.VirtualModule, self).__init__(name=module.__name__)
            if module_name:
                if warn:
                    logger.warning('module_name ignored when instantiated with module.')
                
            if schema_obj_name:
                assert schema_obj_name in module.__dict__, f'schema_obj_name: {schema_obj_name} not found in module.'
                schema_obj = module.__dict__[schema_obj_name]
                assert isinstance(schema_obj, dj.Schema), f'schema object should be of type {dj.Schema} not {type(schema_obj)}.'
            else:
                schemas = {k: {'obj': v, 'database': v.database} for k, v in module.__dict__.items() if isinstance(v, dj.Schema)}
                assert len(schemas.keys())==1, f"Found multiple schema objects with names {list(schemas.keys())}, mapping to respective databases {[v['database'] for v in schemas.values()]}. Specify the name of the schema object to instantiate with arg schema_obj_name." 
                schema_obj = list(schemas.values())[0]['obj']
            
            self.__dict__.update(module.__dict__)
            self.__dict__['schema'] = schema_obj
            
            if spawn_missing_classes:
                schema_obj.spawn_missing_classes(context=self.__dict__)
                
            self.load_dependencies(schema_obj.connection, load_dependencies, verbose=False)
                
            if add_objects:
                self.__dict__.update(add_objects)
        
        else:
            raise ValueError('Provide schema_name or module.')      
        
        if add_externals:
            register_externals(add_externals)
        
        if enable_dj_flags:
            enable_datajoint_flags()
            
        add_datajoint_plus(self)

    def load_dependencies(self, connection, force=True, verbose=True):
        """
        Loads dependencies into DataJoint networkx graph. 
        """
        load_dependencies(connection=connection, force=force, verbose=verbose)


"""
Exception classes for the DataJointPlus
"""
from datajoint.errors import DataJointError


class DataJointPlusError(DataJointError):
    pass


class ValidationError(DataJointPlusError):
    """
    Validation failure.
    """

class OverwriteError(DataJointPlusError):
    """
    Overwrite not permissible.
    """

class MotifError(DataJointPlusError):
    pass


class MakerError(MotifError):
    pass


class MakerInputError(MotifError):
    pass


class MakerMethodError(MotifError):
    pass


class MakerDestinationError(MotifError):
    pass


"""
User tables for DataJointPlus Motifs
"""

from collections import namedtuple
import inspect

import datajoint as dj
import numpy as np

from datajoint_plus.utils import classproperty, format_table_name, wrap, split_full_table_name

from .base import BaseMaster, BasePart
from .motif_base import MotifMaster, Nested

### MASTER

class Entity(MotifMaster, BaseMaster, dj.Lookup):
    @classmethod
    def _init_validation(cls, **kwargs):
        assert getattr(cls, 'hash_name', None) is not None, f'Subclasses of {cls.__base__.__name__} must implement hash_name.'

        if cls._hash_len is None:
            cls._hash_len = 32 # default hash length
        else:
            if not (isinstance(cls.hash_len, int) and (cls._hash_len > 0 and cls._hash_len <= 32)):
                raise NotImplementedError('_hash_len attribute must be an integer within range: [1, 32].')

        super()._init_validation(**kwargs)

    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)

    @classmethod
    def is_entity(cls):
        return issubclass(cls, Entity)    
    
    @classmethod
    def stores(cls):
        stores = [c for c in cls.parts(as_cls=True) if getattr(c, 'is_nested_store', False)]
        store_names = [c.class_name.split('.')[1] for c in stores]
        return namedtuple('NestedStore', store_names)(*stores) 

    @classmethod
    def makers(cls):
        makers = [c for c in cls.parts(as_cls=True) if getattr(c, 'is_nested_maker', False)]
        maker_names = [c.class_name.split('.')[1] for c in makers]
        return namedtuple('NestedStore', maker_names)(*makers) 

    @classmethod
    def get(cls, key, **kwargs):
        return cls.r1p(key, include_parts=cls.stores(), **kwargs).get()
    
    @classmethod
    def get_with_hash(cls, hash, **kwargs):
        return cls.get({cls.hash_name: hash}, **kwargs)

    @classmethod
    def restrict_stores(cls, restr={}, **kwargs):
        return cls.restrict_parts(part_restr=restr, include_parts=cls.stores(), **kwargs)
    
    @classmethod
    def restrict_one_store(cls, restr={}, **kwargs):
        return cls.r1p(part_restr=restr, include_parts=cls.stores(), **kwargs)
        
    @classmethod
    def restrict_one_store_with_hash(cls, hash, hash_name=None, **kwargs):
        return cls.r1pwh(hash=hash, hash_name=hash_name, include_parts=cls.stores(), **kwargs)
    
    r1s = restrict_one_store # alias
    
    r1swh = restrict_one_store_with_hash # alias
    
    @classmethod
    def restrict_makers(cls, restr={}, **kwargs):
        return cls.restrict_parts(part_restr=restr, include_parts=cls.makers(), **kwargs)
        
    @classmethod
    def restrict_one_maker(cls, restr={}, **kwargs):
        return cls.r1p(part_restr=restr, include_parts=cls.makers(), **kwargs)
    
    @classmethod
    def restrict_one_maker_with_hash(cls, hash, hash_name=None, **kwargs):
        return cls.r1pwh(hash=hash, hash_name=hash_name, include_parts=cls.makers(), **kwargs)
    
    r1m = restrict_one_maker # alias 
    
    r1mwh = restrict_one_maker_with_hash # alias


class MethodGroup(MotifMaster, BaseMaster, dj.Lookup):
    @classmethod
    def _init_validation(cls, **kwargs):
        assert getattr(cls, 'hash_name', None) is not None, f'Subclasses of {cls.__base__.__name__} must implement hash_name.'

        if cls._hash_len is None:
            cls._hash_len = 32 # default hash length
        else:
            if not (isinstance(cls.hash_len, int) and (cls._hash_len > 0 and cls._hash_len <= 32)):
                raise NotImplementedError('_hash_len attribute must be an integer within range: [1, 32].')

        super()._init_validation(**kwargs)
    
    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)

    @classmethod
    def is_method_group(cls):
        return issubclass(cls, MethodGroup)

class DestinationGroup(MotifMaster, BaseMaster, dj.Lookup):    
    @classmethod
    def _init_validation(cls, **kwargs):
        assert getattr(cls, 'hash_name', None) is not None, f'Subclasses of {cls.__base__.__name__} must implement hash_name.'

        if cls._hash_len is None:
            cls._hash_len = 32 # default hash length
        else:
            if not (isinstance(cls.hash_len, int) and (cls._hash_len > 0 and cls._hash_len <= 32)):
                raise NotImplementedError('_hash_len attribute must be an integer within range: [1, 32].')

        super()._init_validation(**kwargs)
        
    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)

    @classmethod
    def is_destination_group(cls):
        return issubclass(cls, DestinationGroup)
    
    def update(self):
        [p().update() for p in self.parts(as_cls=True) if getattr(p, 'is_nested_destination', False)]


### NESTED
class NestedMethod(Nested, BasePart, dj.Part):
    @classmethod
    def is_nested_method(cls):
        return issubclass(cls, NestedMethod)    
        
    @classmethod
    def _init_validation(cls, **kwargs):
        if not getattr(cls, 'enable_hashing', False):
            raise NotImplementedError(f'Subclasses of {cls.__base__.__name__} must have enable_hashing set to True.')

        super()._init_validation(**kwargs)

    def __init_subclass__(cls, **kwargs):
        cls.enable_hashing = True
        cls._init_validation(**kwargs)


class NestedStore(Nested, BasePart, dj.Part):
    @classmethod
    def _init_validation(cls, **kwargs):
        super()._init_validation(**kwargs)

    def __init_subclass__(cls, **kwargs):
        cls._init_validation(**kwargs)

    @classmethod
    def is_nested_store(cls):
        return issubclass(cls, NestedStore)

    @classmethod
    def put(cls, result):
        cls.insert1(result, ignore_extra_fields=True, insert_to_master=True)


class NestedDestination(Nested, BasePart, dj.Part):
    @classmethod
    def _init_validation(cls, **kwargs):
        super()._init_validation(**kwargs)
        
    def __init_subclass__(cls, **kwargs):
        cls.enable_hashing = True
        cls.hash_group = True
        cls.hashed_attrs = 'table_id'
        cls.definition = ''.join([
        f"""
        # {cls().__class__.__name__}
        -> master
        ---
        table_id: varchar(32) # id of table in schema.tables
        """
        ])
        cls._init_validation(**kwargs)
        

    @classmethod
    def is_nested_destination(cls):
        return issubclass(cls, NestedDestination)
    
    @classmethod
    def add_destination(cls, table_ids):
        cls.insert([{'table_id': t} for t in wrap(table_ids)], ignore_extra_fields=True, skip_duplicates=True, insert_to_master=True)

    def update(self):
        if self.destinations is not None:
            for table_id in wrap(self.destinations):
                self.add_destination(table_id)

    def put(self, result, key=None):
        if key is None:
            self.goto().put(result)
        else:
            (self & key).goto().put(result)


class NestedMaker(Nested, BasePart, dj.Part, dj.Computed): 
    _upstream = None
    _method = None
    _destination = None

    @classmethod
    def is_nested_maker(cls):
        return issubclass(cls, NestedMaker)

    @classmethod
    def _init_validation(cls, **kwargs):
        # Single item
        for name in ['_method', '_destination']:
            item = getattr(cls, name, None)
            if isinstance(item, list) or isinstance(item, tuple):
                if len(item) > 1:
                    raise NotImplementedError(f'Only one {name[1:]} allowed in {name}.')
                else:
                    setattr(cls, name, item[0])

        if not getattr(cls, 'enable_hashing', False):
            raise NotImplementedError(f'Subclasses of {cls.__base__.__name__} must have enable_hashing set to True.')

        for required in ['hash_name', '_method', '_destination']:
            if getattr(cls, required, None) is None:
                raise NotImplementedError(f'Subclasses of {cls.__base__.__name__} must specify {required}')
        
        # Make sure _upstream wrapped in list or tuple
        for name in ['_upstream']:
            item = getattr(cls, name, None)
            if item is not None:
                setattr(cls, name, wrap(item))

    def __init_subclass__(cls, **kwargs):
        cls.enable_hashing = True
        cls._init_validation(**kwargs)

        for name in ['upstream', 'method', 'destination']:
            setattr(cls, name, classproperty(getattr(cls, name)))

        if getattr(cls, 'hashed_attrs', None) is None:
            cls.hashed_attrs = cls.key_source.primary_key
        if getattr(cls, '_upstream_definition', None) is None:
            cls._upstream_definition = ''.join([f'-> {u.class_name} \n' for u in cls._upstream]) if cls._upstream is not None else ''
        if getattr(cls, '_method_definition', None) is None:
            cls._method_definition = ''.join(['-> ', cls._method.class_name, '\n'])
        if getattr(cls, '_destination_definition', None) is None:
            cls._destination_definition = ''.join(['-> ', cls._destination.class_name, '\n'])
        if getattr(cls, 'additional_pk_definition', None) is None:
            cls._additional_pk_definition = ''
        if getattr(cls, 'additional_sk_definition', None) is None:
            cls._additional_sk_definition = ''

        cls.definition = ''.join([
            f"""
            # {cls().__class__.__name__}
            -> master
            """,
            cls._upstream_definition,
            cls._method_definition,
            cls._additional_pk_definition,
            """
            ---
            """,
            cls._destination_definition,
            cls._additional_sk_definition,
            """
            ts_inserted=CURRENT_TIMESTAMP: timestamp #
            """
        ])

        super()._init_validation(**kwargs)

    def upstream(cls):
        if cls._upstream is not None:
            names = []
            for up in cls._upstream:
                try:
                    names.append(up.class_name)
                except AttributeError:
                    names.append(format_table_name(split_full_table_name(up.from_clause)[1]) + up.__class__.__name__)

            return namedtuple('Upstream', *names)(*cls._upstream)

    def method(cls):
        return cls._method
    
    def destination(cls):
        return cls._destination

    @property
    def key_source(self):
        if self.upstream is not None:
            return np.product(*self.upstream) * self.method
        else:
            return self.method

    def make(self, key):
        inputs = {}
        if self.upstream is not None:
            for u in self.upstream:
                if inspect.isclass(u):
                    u = u()
                if getattr(self, 'get_upstream_keys_only', False):
                    inputs.update(**(u & key).fetch1('KEY'))
                else:
                    try:
                        inputs.update(**u.get(key))
                    except:
                        inputs.update(**(u & key).fetch1())

        if getattr(self.method, 'is_method_group', False):
            result = self.method.r1p(key).run(inputs)
        elif getattr(self.method, 'is_nested_method', False):
            result = (self.method & key).run(inputs)
        else:
            raise AttributeError(f'table type {self.method.__class__.__base__.__name__} not supported.')
            
        result.update({self.hash_name: self.hash1(key)})
        
        if inspect.isclass(self.destination):
            self.destination = self.destination()
        self.destination.update()
        dkey = self.destination.fetch1('KEY')
        if getattr(self.destination, 'is_destination_group', False):
            destination = self.destination.r1p(dkey)
        elif getattr(self.destination, 'is_nested_destination', False):
            destination = (self.destination & dkey)
        else:
            raise AttributeError(f'table type {self.destination.__class__.__base__.__name__} not a supported.')
        
        key.update(**dkey)
        destination.put(result)
        self.insert1(key)


# class NestedMethodAdapter(Nested, BasePart, dj.Part, dj.Imported):
#     _upstream = None
#     _method = None
#     _destination = None
    
#     @classmethod
#     def _init_validation(cls, **kwargs):
#         for required in ['_method']:
#             if getattr(cls, required) is None:
#                 raise NotImplementedError(f'Subclasses of NestedCompute must specify {required}')

#         for name in ['_upstream', '_destination']:
#             if getattr(cls, name) is not None:
#                 raise NotImplementedError(f'Subclasses of NestedMethodAdapter must not contain {name}.')
        
#         if isinstance(cls._method, list) or isinstance(cls._method, tuple):
#             if len(cls._method) > 1:
#                 raise NotImplementedError(f'Only one class allowed in method.')
#             else:
#                 cls._method = cls._method[0]
        
#     def __init_subclass__(cls, **kwargs):
#         cls.hashed_attrs = cls.key_source.heading.primary_key
#         cls.definition = ''.join([
#             f"""
#             # {cls().__class__.__name__}
#             -> master
#             """,
#             ''.join(['->', cls._method.class_name, '\n']),
#             """
#             ---
#             ts_inserted=CURRENT_TIMESTAMP: timestamp #
#             """
#         ])
        
#         cls._init_validation(**kwargs)

#         @classproperty
#         def method(cls):
#             return cls._method

#         @classproperty
#         def key_source(cls):
#             return cls.method
    
#     def run(self, **kwargs):
#         return self.method.r1p(self).run(**kwargs)
